# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2019-2024 SRON - Netherlands Institute for Space Research
#   All Rights Reserved
#
# License:  BSD-3-Clause
"""Necessary but empty file."""

from __future__ import annotations

__all__ = ["pyspex_version"]

import contextlib
from importlib.metadata import PackageNotFoundError, version


def pyspex_version(full: bool = False, githash: bool = False) -> str:
    """Return the software version as obtained from git."""
    with contextlib.suppress(PackageNotFoundError):
        __version__ = version("pyspex")

    if full:
        return __version__

    if githash:
        res = __version__.split("+g")
        if len(res) > 1:
            return res[1].split(".")[0]

        return "v" + "".join([f"{int(x):02d}" for x in res[0].split(".")])

    return __version__.split("+")[0]



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
"""The Sphinx configuration file for the package pySpex."""

import os
import sys
from importlib import metadata

sys.path.insert(0, os.path.abspath("../src"))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "pyspex"
copyright = "2022, SRON"
author = "Richard van Hees"
release = metadata.version("pyspex").split("+")[0]

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.doctest",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]



#
# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2022-2024 SRON - Netherlands Institute for Space Research
#    All Rights Reserved
#
# License:  BSD-3-Clause
"""Check validity and format of SPEXone level-0 products."""

from __future__ import annotations

__all__ = ["check_input_files"]

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from dataclasses import dataclass


def check_input_files(config: dataclass) -> dataclass:
    """Check SPEXone level-0 files on existence and format.

    Parameters
    ----------
    config :  dataclass
       Dataclass that contains the settings of the L0-L1A processor

    Returns
    -------
    dataclass
       fields 'l0_format' {'raw', 'st3', 'dsb'} and 'l0_list' are updated.

    Raises
    ------
    FileNotFoundError
       If files are not found on the system.
    TypeError
       If determined file type differs from value supplied by user.

    """
    file_list = config.l0_list
    if file_list[0].suffix == ".H":
        if not file_list[0].is_file():
            raise FileNotFoundError(file_list[0])
        data_dir = file_list[0].parent
        file_stem = file_list[0].stem
        file_list = (
            sorted(data_dir.glob(file_stem + ".[0-9]"))
            + sorted(data_dir.glob(file_stem + ".?[0-9]"))
            + sorted(data_dir.glob(file_stem + "_hk.[0-9]"))
        )
        if not file_list:
            raise FileNotFoundError(file_stem + ".[0-9]")

        config.l0_format = "raw"
        config.l0_list = file_list
    elif file_list[0].suffix == ".ST3":
        if not file_list[0].is_file():
            raise FileNotFoundError(file_list[0])
        config.l0_format = "st3"
        config.l0_list = [file_list[0]]
    elif file_list[0].suffix == ".spx":
        file_list_out = []
        for flname in file_list:
            if not flname.is_file():
                raise FileNotFoundError(flname)

            if flname.suffix == ".spx":
                file_list_out.append(flname)

        if not file_list_out:
            raise FileNotFoundError(file_list)
        config.l0_format = "dsb"
        config.l0_list = file_list_out
    else:
        raise TypeError("Input files not recognized as SPEXone level-0 data")

    return config



#
# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2022-2024 SRON - Netherlands Institute for Space Research
#    All Rights Reserved
#
# License:  BSD-3-Clause
"""`SPXtlm`, class to read/access PACE/SPEXone telemetry data."""

from __future__ import annotations

__all__ = ["SPXtlm"]

import datetime as dt
import logging
from copy import copy
from pathlib import Path
from typing import TYPE_CHECKING

import h5py
import numpy as np

# pylint: disable=no-name-in-module
from netCDF4 import Dataset

from .hkt_io import HKTio
from .l1a_io import L1Aio
from .lib import pyspex_version
from .lib.hk_tlm import HKtlm
from .lib.leap_sec import get_leap_seconds
from .lib.sci_tlm import DET_CONSTS, SCItlm
from .lv0_lib import dump_hkt, dump_science, read_lv0_data

if TYPE_CHECKING:
    from dataclasses import dataclass

# - global parameters -----------------------
module_logger = logging.getLogger("pyspex.tlm")

FULLFRAME_BYTES = 2 * DET_CONSTS["dimFullFrame"]
DATE_MIN = dt.datetime(2020, 1, 1, 1, tzinfo=dt.UTC)
TSTAMP_MIN = int(DATE_MIN.timestamp())


# - helper functions ------------------------
def get_l1a_filename(
    config: dataclass,
    coverage: tuple[dt.datetime, dt.datetime],
    mode: str | None = None,
) -> Path:
    """Return filename of level-1A product.

    Parameters
    ----------
    config :  dataclass
        Settings for the L0->l1A processing
    coverage :  tuple[dt.datetime, dt.datetime]
        time_coverage_start and time_coverage_end of the data
    mode :  {'binned', 'full'} | None, default=None
        Select Science packages with full-frame images or binned images

    Returns
    -------
    Path
        Filename of level-1A product.

    Notes
    -----
    === Inflight ===
    L1A file name format, following the NASA ... naming convention:
        PACE_SPEXONE[_TTT].YYYYMMDDTHHMMSS.L1A[.Vnn].nc
    where
        TTT is an optional data type (e.g., for the calibration data files)
        YYYYMMDDTHHMMSS is time stamp of the first image in the file
        Vnn file-version number (omitted when nn=1)

    for example (file-version=1):
        [Science Product] PACE_SPEXONE.20230115T123456.L1A.nc
        [Calibration Product] PACE_SPEXONE_CAL.20230115T123456.L1A.nc
        [Dark science Product] PACE_SPEXONE_DARK.20230115T123456.L1A.nc

    === OCAL ===
    L1A file name format:
        SPX1_OCAL_<msm_id>[_YYYYMMDDTHHMMSS]_L1A_vvvvvvv.nc
    where
        msm_id is the measurement identifier
        YYYYMMDDTHHMMSS is time stamp of the first image in the file
        vvvvvvv is the git-hash string of the pyspex repository

    """
    if config.outfile:
        return config.outdir / config.outfile

    # +++++ in-flight product-name convention +++++
    if config.l0_format != "raw":
        if config.eclipse is None:
            subtype = "_OCAL"
        elif not config.eclipse:
            subtype = ""
        else:
            subtype = "_CAL" if mode == "full" else "_DARK"

        prod_ver = "" if config.file_version == 1 else f".V{config.file_version:02d}"

        return config.outdir / (
            f'PACE_SPEXONE{subtype}'
            f'.{coverage[0].strftime("%Y%m%dT%H%M%S"):15s}'
            f'.L1A{prod_ver}.nc'
        )

    # +++++ OCAL product-name convention +++++
    # determine measurement identifier
    msm_id = config.l0_list[0].stem
    try:
        new_date = dt.datetime.strptime(msm_id[-22:], "%y-%j-%H:%M:%S.%f").strftime(
            "%Y%m%dT%H%M%S.%f"
        )
    except ValueError:
        pass
    else:
        msm_id = msm_id[:-22] + new_date

    return config.outdir / f"SPX1_OCAL_{msm_id}_L1A_{pyspex_version(githash=True)}.nc"


def add_proc_conf(l1a_file: Path, yaml_conf: Path) -> None:
    """Add dataset 'processor_configuration' to an existing L1A product.

    Parameters
    ----------
    l1a_file :  Path
       name of an existing L1A product.
    yaml_conf :  Path
       name of the YAML file with the processor settings

    """
    with Dataset(l1a_file, "r+") as fid:
        dset = fid.createVariable("processor_configuration", str)
        dset.comment = (
            "Configuration parameters used during"
            " the processor run that produced this file."
        )
        dset.markup_language = "YAML"
        dset[0] = "".join(
            [
                s
                for s in yaml_conf.open(encoding="ascii").readlines()
                if not (s == "\n" or s.startswith("#"))
            ]
        )


# - class SPXtlm ----------------------------
class SPXtlm:
    """Access/convert parameters of SPEXone Science telemetry data.

    Notes
    -----
    This class has the following methods::

     - set_coverage(coverage: tuple[datetime, datetime] | None,
                    update: boot = False) -> None
     - reference_date() -> datetime
     - time_coverage_start() -> datetime
     - time_coverage_end() -> datetime
     - from_hkt(flnames: str | Path | list[Path], *,
                instrument: str | None = None, dump: bool = False) -> None
     - from_lv0(flnames: str | Path | list[Path], file_format: str = 'dsb',
                *, tlm_type: str | None = None,
                debug: bool = False, dump: bool = False) -> None
     - from_l1a(flname: str | Path,
                *, tlm_type: str | None = None,
                mps_id: int | None = None) -> None
     - gen_l1a(config: dataclass) -> None


    Examples
    --------
    Read data from SPEXone level-0 products

    >>> from pyspex.tlm import SPXtlm
    >>> spx = SPXtlm()
    >>> spx.from_l0(list_l0_products, tlm_type='hk')
    # returns list of TcAccept, TcReject, TcExecute, TcFail and EventRp
    >>> spx.nomhk.events
    # return detector images
    >>> spx.from_l0(list_l0_products, tlm_type='sci')
    >>> spx.science.images

    Read data from a SPEXone level-1a product

    >>> from pyspex.tlm import SPXtlm
    >>> spx = SPXtlm()
    >>> spx.from_l1a(l1a_product, tlm_type='sci', mps_is=47)
    >>> spx.science.images

    Read data from a PACE-HKT product

    >>> from pyspex.tlm import SPXtlm
    >>> spx = SPXtlm()
    >>> spx.from_hkt(hkt_product)
    # returns list of TcAccept, TcReject, TcExecute, TcFail and EventRp
    >>> spx.nomhk.events

    """

    def __init__(self: SPXtlm) -> None:
        """Initialize SPXtlm object."""
        self.mode = "all"
        self.logger = logging.getLogger(__name__)
        self.file_list: list[Path] | None = None
        self._coverage: list[dt.datetime, dt.datetime] | None = None
        self.nomhk = HKtlm()
        self.science = SCItlm()

    @property
    def coverage(self: SPXtlm) -> list[dt.datetime, dt.datetime] | None:
        """Return data time_coverage."""
        return self._coverage

    def set_coverage(
        self: SPXtlm,
        coverage_new: list[dt.datetime, dt.datetime] | None,
        extent: bool = False,
    ) -> None:
        """Set, reset or update the class attribute `coverage`.

        Parameters
        ----------
        coverage_new :  list[dt.datetime, dt.datetime] | None,
           Provide new time_coverage range, or reset when coverage_new is None
        extent :  bool, default=False
           Extent existing time_coverage range with coverage_new

        Examples
        --------
        Reset self.coverage to a new time_coverage:
        > self.set_coverage(None)
        > self.set_coverage(coverage_new)

        Extent self.coverage with a new time_coverage
        if self.coverage is not None else self.coverage will be coverage_new:
        > self.set_coverage(coverage_new, extent=True)

        This will not change self.coverage to coverage_new
        when self.coverage is not None:
        > self.set_coverage(coverage_new)

        """
        if self._coverage is None or coverage_new is None:
            self._coverage = coverage_new
            return

        if not extent:
            return

        one_hour = dt.timedelta(hours=1)
        if self._coverage[0] - one_hour < coverage_new[0] < self._coverage[0]:
            self._coverage[0] = coverage_new[0]
        if self._coverage[1] < coverage_new[1] < self._coverage[1] + one_hour:
            self._coverage[1] = coverage_new[1]

    @property
    def reference_date(self: SPXtlm) -> dt.datetime:
        """Return date of reference day (tzone aware)."""
        if self._coverage is None:
            raise ValueError("no valid timestamps found")

        return dt.datetime.combine(self._coverage[0].date(), dt.time(0), dt.UTC)

    @property
    def time_coverage_start(self: SPXtlm) -> dt.datetime | None:
        """Return time_coverage_start."""
        return None if self._coverage is None else self._coverage[0]

    @property
    def time_coverage_end(self: SPXtlm) -> dt.datetime | None:
        """Return time_coverage_end."""
        return None if self._coverage is None else self._coverage[1]

    def sel(self: SPXtlm, mask: np.NDArray[bool]) -> SPXtlm:
        """Return subset of SPXtlm object using a mask array."""
        spx = copy(self)
        spx.set_coverage(None)
        spx.science = self.science.sel(mask)

        # set Science time_coverage_range
        indices = mask.nonzero()[0]
        if len(indices) == 1:
            frame_period = dt.timedelta(milliseconds=self.science.frame_period(0))
            spx.set_coverage(
                [
                    spx.science.tstamp[0]["dt"],
                    spx.science.tstamp[0]["dt"] + frame_period,
                ]
            )
        else:
            frame_period = dt.timedelta(milliseconds=self.science.frame_period(-1))
            spx.set_coverage(
                [
                    spx.science.tstamp[0]["dt"],
                    spx.science.tstamp[-1]["dt"] + frame_period,
                ]
            )
        # select nomhk data within Science time_coverage_range
        hk_tstamps = np.array(
            [x.replace(tzinfo=None) for x in self.nomhk.tstamp], dtype="datetime64"
        )
        dt_min = np.datetime64(spx.coverage[0].replace(tzinfo=None)) - np.timedelta64(
            1, "s"
        )
        dt_max = np.datetime64(spx.coverage[1].replace(tzinfo=None)) + np.timedelta64(
            1, "s"
        )
        spx.nomhk = self.nomhk.sel((hk_tstamps >= dt_min) & (hk_tstamps <= dt_max))
        return spx

    def from_hkt(
        self: SPXtlm,
        flnames: str | Path | list[Path],
        *,
        instrument: str | None = None,
        dump: bool = False,
    ) -> None:
        """Read telemetry data from PACE HKT product(s).

        Parameters
        ----------
        flnames :  str | Path | list[Path]
           sorted list of PACE_HKT filenames (netCDF4 format)
        instrument :  {'spx', 'sc', 'oci', 'harp'}, default='spx'
           abbreviations for the PACE instruments
        dump :  bool, default=False
           dump header information of the telemetry packages @1Hz for
           debugging purposes

        """
        self.set_coverage(None)
        if isinstance(flnames, str | Path):
            flnames = [flnames] if isinstance(flnames, Path) else [Path(flnames)]
        if instrument is None:
            instrument = "spx"
        elif instrument not in ["spx", "sc", "oci", "harp"]:
            raise KeyError("instrument not in ['spx', 'sc', 'oci', 'harp']")

        self.file_list = flnames

        # check number of telemetry data-packages
        ccsds_hk = HKTio(flnames).housekeeping(instrument)
        if not ccsds_hk:
            return

        # perform dump of telemetry data-packages
        if dump:
            dump_hkt(flnames[0].stem + "_hkt.dump", ccsds_hk)
            return

        # check if TAI timestamp is valid
        ii = len(ccsds_hk) // 2
        if (tai_sec := ccsds_hk[ii]["hdr"]["tai_sec"][0]) == 0:
            return

        # set epoch
        epoch = dt.datetime(1958, 1, 1, tzinfo=dt.UTC)
        epoch -= dt.timedelta(seconds=get_leap_seconds(float(tai_sec)))
        self.nomhk.extract_l0_hk(ccsds_hk, epoch)

        # reject nomHK records before or after a big time-jump
        _mm = np.diff(self.nomhk.tstamp) > np.timedelta64(1, "D")
        if np.any(_mm):
            indx = _mm.nonzero()[0]
            _mm = np.full(self.nomhk.size, True, dtype=bool)
            _mm[indx[0] + 1 :] = False
            if np.sum(_mm) < self.nomhk.size // 2:
                _mm = ~_mm
            self.logger.warning(
                "rejected nomHK: %d -> %d", self.nomhk.size, np.sum(_mm)
            )
            self.nomhk.sel(_mm)

        # set time-coverage
        self.set_coverage(
            [self.nomhk.tstamp[0], self.nomhk.tstamp[-1] + dt.timedelta(seconds=1)]
        )

    def from_lv0(
        self: SPXtlm,
        flnames: str | Path | list[Path],
        file_format: str = "dsb",
        *,
        tlm_type: str | None = None,
        debug: bool = False,
        dump: bool = False,
    ) -> None:
        """Read telemetry data from SPEXone level-0 product(s).

        Parameters
        ----------
        flnames :  str | Path | list[Path]
           sorted list of CCSDS filenames
        file_format : {'raw', 'st3', 'dsb'}, default='dsb'
           type of CCSDS data
        tlm_type :  {'hk', 'sci', 'all'}, default='all'
           select type of telemetry packages
           Note that we always read the complete level-0 products.
        debug : bool, default=False
           run in debug mode, read only packages heades
        dump :  bool, default=False
           dump header information of the telemetry packages @1Hz for
           debugging purposes

        """
        self.set_coverage(None)
        if isinstance(flnames, str | Path):
            flnames = [flnames] if isinstance(flnames, Path) else [Path(flnames)]
        if file_format not in ["raw", "st3", "dsb"]:
            raise KeyError("file_format not in ['raw', 'st3', 'dsb']")
        if tlm_type is None:
            tlm_type = "all"
        elif tlm_type not in ["hk", "sci", "all"]:
            raise KeyError("tlm_type not in ['hk', 'sci', 'all']")
        self.file_list = flnames

        # read telemetry data
        ccsds_sci, ccsds_hk = read_lv0_data(flnames, file_format, debug=debug)

        # perform ASCII dump
        if dump and ccsds_hk:
            dump_hkt(flnames[0].stem + "_hkt.dump", ccsds_hk)
        if dump and ccsds_sci:
            dump_science(flnames[0].stem + "_sci.dump", ccsds_sci)

        # exit when debugging or only an ASCII data-dump is requested
        if debug or dump:
            return

        # exit when Science data is requested and no Science data is available
        #   or when housekeeping is requested and no housekeeping data is available
        if (not ccsds_sci and tlm_type == "sci") or (not ccsds_hk and tlm_type == "hk"):
            self.logger.info("Asked for tlm_type=%s, but none found", tlm_type)
            return

        # set epoch
        if file_format == "dsb":
            # check if TAI timestamp is valid
            ii = len(ccsds_hk) // 2
            if (tai_sec := ccsds_hk[ii]["hdr"]["tai_sec"][0]) == 0:
                return

            epoch = dt.datetime(1958, 1, 1, tzinfo=dt.UTC)
            epoch -= dt.timedelta(seconds=get_leap_seconds(float(tai_sec)))
        else:
            epoch = dt.datetime(1970, 1, 1, tzinfo=dt.UTC)

        if tlm_type != "hk":
            # collect Science telemetry data
            if self.science.extract_l0_sci(ccsds_sci, epoch) == 0:
                self.logger.info("no valid Science package found")
            else:
                # reject Science records before or after a big time-jump
                _mm = np.diff(self.science.tstamp["dt"]) > np.timedelta64(1, "D")
                if np.any(_mm):
                    indx = _mm.nonzero()[0]
                    _mm = np.full(self.science.size, True, dtype=bool)
                    _mm[indx[0] + 1 :] = False
                    if np.sum(_mm) < self.science.size // 2:
                        _mm = ~_mm
                    self.logger.warning(
                        "rejected science: %d -> %d", self.science.size, np.sum(_mm)
                    )
                    self.science.sel(_mm)

                # set time-coverage
                intg = dt.timedelta(milliseconds=self.science.frame_period(-1))
                self.set_coverage(
                    [self.science.tstamp["dt"][0], self.science.tstamp["dt"][-1] + intg]
                )

        # collected NomHK telemetry data
        if tlm_type != "sci" and ccsds_hk:
            self.nomhk.extract_l0_hk(ccsds_hk, epoch)

            # reject nomHK records before or after a big time-jump
            _mm = np.diff(self.nomhk.tstamp) > np.timedelta64(1, "D")
            if np.any(_mm):
                indx = _mm.nonzero()[0]
                _mm = np.full(self.nomhk.size, True, dtype=bool)
                _mm[indx[0] + 1 :] = False
                if self.coverage is None:
                    if np.sum(_mm) < self.nomhk.size // 2:
                        _mm = ~_mm
                else:
                    dt_start = abs(self.coverage[0] - self.nomhk.tstamp[0])
                    dt_end = abs(self.coverage[-1] - self.nomhk.tstamp[-1])
                    if dt_start > dt_end:
                        _mm = ~_mm
                self.logger.warning(
                    "rejected nomHK: %d -> %d", self.nomhk.size, np.sum(_mm)
                )
                self.nomhk.sel(_mm)

            # set time-coverage (only, when self._coverage is None)
            self.set_coverage(
                [self.nomhk.tstamp[0], self.nomhk.tstamp[-1] + dt.timedelta(seconds=1)]
            )

    def from_l1a(
        self: SPXtlm,
        flname: str | Path,
        *,
        tlm_type: str | None = None,
        mps_id: int | None = None,
    ) -> None:
        """Read telemetry data from SPEXone level-1A product.

        Parameters
        ----------
        flname :  Path
           name of one SPEXone level-1A product
        tlm_type :  {'hk', 'sci', 'all'}, default='all'
           select type of telemetry packages
        mps_id :  int, optional
           select on MPS ID

        """
        self.set_coverage(None)
        if tlm_type is None:
            tlm_type = "all"
        elif tlm_type not in ["hk", "sci", "all"]:
            raise KeyError("tlm_type not in ['hk', 'sci', 'all']")

        self.file_list = [flname]
        with h5py.File(flname) as fid:
            # collect Science telemetry data
            if tlm_type != "hk":
                self.science.extract_l1a_sci(fid, mps_id)
                if self.science.size == 0:
                    return

                _mm = self.science.tstamp["tai_sec"] > TSTAMP_MIN
                if np.any(_mm):
                    tstamp = self.science.tstamp["dt"][_mm]
                    ii = int(np.nonzero(_mm)[0][-1])
                    intg = dt.timedelta(milliseconds=self.science.frame_period(ii))
                    self.set_coverage([tstamp[0], tstamp[-1] + intg])

            # collected NomHk telemetry data
            if tlm_type != "sci":
                self.nomhk.extract_l1a_hk(fid, mps_id)
                if self.nomhk.size == 0:
                    return
                self.set_coverage(
                    [
                        self.nomhk.tstamp[0],
                        self.nomhk.tstamp[-1] + dt.timedelta(seconds=1),
                    ]
                )

    def full(self: SPXtlm) -> SPXtlm:
        """Select full-frame measurements."""
        self.mode = "full"
        if self.science.size == 0:
            return self

        sci_mask = (self.science.tstamp["tai_sec"] > TSTAMP_MIN) & (
            self.science.tlm["IMRLEN"] == FULLFRAME_BYTES
        )
        if np.all(~sci_mask):
            return SPXtlm()  # return empy object
        if np.all(sci_mask):
            return self  # return original object
        self.logger.debug("Rejected %d binned Science images", np.sum(~sci_mask))

        return copy(self).sel(sci_mask)

    def binned(self: SPXtlm) -> SPXtlm:
        """Select binned images from data."""
        self.mode = "binned"
        if self.science.size == 0:
            return self

        sci_mask = (self.science.tstamp["tai_sec"] > TSTAMP_MIN) & (
            self.science.tlm["IMRLEN"] < FULLFRAME_BYTES
        )
        if np.all(~sci_mask):
            return SPXtlm()  # return empy object
        if np.all(sci_mask):
            return self  # return original object
        self.logger.debug("Rejected %d full-frame Science images", np.sum(~sci_mask))

        return copy(self).sel(sci_mask)

    def gen_l1a(self: SPXtlm, config: dataclass) -> None:
        """Generate a SPEXone level-1A product.

        Parameters
        ----------
        config :  dataclass
           Settings for the L0->L1A processing

        """
        # sanity check
        if self.science.size == 0 and self.nomhk.size == 0:
            return

        if self.science.size > 0:
            mps_list = np.unique(self.science.tlm["MPS_ID"])
            self.logger.debug("unique Science MPS: %s", mps_list)
            self.nomhk = self.nomhk.sel(np.in1d(self.nomhk.tlm["MPS_ID"], mps_list))

        dims = {
            "number_of_images": self.science.size,
            "samples_per_image": (
                DET_CONSTS["dimRow"]
                if self.science.size == 0
                else np.max(self.science.tlm["IMRLEN"]) // 2
            ),
            "hk_packets": self.nomhk.size,
        }
        l1a_path = get_l1a_filename(config, self.coverage, self.mode)
        with L1Aio(
            l1a_path,
            self.reference_date,
            dims,
            compression=config.compression,
        ) as l1a:
            l1a.fill_global_attrs(inflight=config.l0_format != "raw")
            if config.processing_version:
                l1a.set_attr("processing_version", config.processing_version)
            l1a.set_attr("icu_sw_version", f'0x{self.nomhk.tlm["ICUSWVER"][0]:x}')
            l1a.set_attr(
                "time_coverage_start",
                self.time_coverage_start.isoformat(timespec="milliseconds"),
            )
            l1a.set_attr(
                "time_coverage_end",
                self.time_coverage_end.isoformat(timespec="milliseconds"),
            )
            l1a.set_attr("input_files", [x.name for x in config.l0_list])
            self.logger.debug("(1) initialized level-1A product")

            self._fill_engineering(l1a)
            self.logger.debug("(2) added engineering data")
            self._fill_science(l1a, dims["samples_per_image"])
            self.logger.debug("(3) added science data")
            self._fill_image_attrs(l1a, config.l0_format)
            self.logger.debug("(4) added image attributes")

        # add PACE navigation information from HKT products
        if config.hkt_list:
            hkt = HKTio(config.hkt_list)
            hkt.navigation()
            hkt.add_nav(l1a_path, self.coverage)
            self.logger.debug("(5) added PACE navigation data")

        # add processor_configuration
        if config.yaml_fl:
            add_proc_conf(l1a_path, config.yaml_fl)

        self.logger.info("successfully generated: %s", l1a_path.name)

    def _fill_engineering(self: SPXtlm, l1a: L1Aio) -> None:
        """Fill datasets in group '/engineering_data'."""
        if self.nomhk.size == 0:
            return

        l1a.set_dset("/engineering_data/NomHK_telemetry", self.nomhk.tlm)
        ref_date = self.reference_date
        l1a.set_dset(
            "/engineering_data/HK_tlm_time",
            [(x - ref_date).total_seconds() for x in self.nomhk.tstamp],
        )
        l1a.set_dset(
            "/engineering_data/temp_detector",
            self.nomhk.convert("TS1_DEM_N_T"),
        )
        l1a.set_dset(
            "/engineering_data/temp_housing",
            self.nomhk.convert("TS2_HOUSING_N_T"),
        )
        l1a.set_dset(
            "/engineering_data/temp_radiator",
            self.nomhk.convert("TS3_RADIATOR_N_T"),
        )

    def _fill_science(self: SPXtlm, l1a: L1Aio, samples_per_image: int) -> None:
        """Fill datasets in group '/science_data'."""
        if self.science.size == 0:
            return

        if len(np.unique([img.size for img in self.science.images])) == 1:
            l1a.set_dset("/science_data/detector_images", self.science.images)
        else:
            images = np.zeros((self.science.size, samples_per_image), dtype="u2")
            for ii, img in enumerate(self.science.images):
                images[ii, : img.size] = img
            l1a.set_dset("/science_data/detector_images", images)
        l1a.set_dset("/science_data/detector_telemetry", self.science.tlm)

    def _fill_image_attrs(self: SPXtlm, l1a: L1Aio, lv0_format: str) -> None:
        """Fill datasets in group '/image_attributes'."""
        if self.science.size == 0:
            return

        l1a.set_dset("/image_attributes/icu_time_sec", self.science.tstamp["tai_sec"])
        # modify attribute units for non-DSB products
        if lv0_format != "dsb":
            # timestamp of 2020-01-01T00:00:00+00:00
            l1a.set_attr(
                "valid_min",
                np.uint32(1577836800),
                ds_name="/image_attributes/icu_time_sec",
            )
            # timestamp of 2024-01-01T00:00:00+00:00
            l1a.set_attr(
                "valid_max",
                np.uint32(1704067200),
                ds_name="/image_attributes/icu_time_sec",
            )
            l1a.set_attr(
                "units",
                "seconds since 1970-01-01 00:00:00",
                ds_name="/image_attributes/icu_time_sec",
            )
        l1a.set_dset(
            "/image_attributes/icu_time_subsec",
            self.science.tstamp["sub_sec"],
        )
        ref_date = self.reference_date
        l1a.set_dset(
            "/image_attributes/image_time",
            [(x - ref_date).total_seconds() for x in self.science.tstamp["dt"]],
        )
        l1a.set_dset(
            "/image_attributes/image_ID",
            np.bitwise_and(self.science.hdr["sequence"], 0x3FFF),
        )
        l1a.set_dset("/image_attributes/binning_table", self.science.binning_table())
        l1a.set_dset("/image_attributes/digital_offset", self.science.digital_offset())
        l1a.set_dset(
            "/image_attributes/exposure_time",
            self.science.exposure_time() / 1000,
        )
        l1a.set_dset(
            "/image_attributes/nr_coadditions",
            self.science.tlm["REG_NCOADDFRAMES"],
        )



#
# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2023-2024 SRON - Netherlands Institute for Space Research
#    All Rights Reserved
#
# License:  BSD-3-Clause
#
"""`SCItlm`, class to read/access PACE/SPEXone telemetry data."""

from __future__ import annotations

__all__ = ["SCItlm"]

import datetime as dt
import logging
from copy import copy
from typing import TYPE_CHECKING

import numpy as np

from .ccsds_hdr import CCSDShdr
from .leap_sec import get_leap_seconds
from .tlm_utils import convert_hk

if TYPE_CHECKING:
    import h5py
    import numpy.typing as npt


# - global parameters -----------------------
DET_CONSTS = {
    "dimRow": 2048,
    "dimColumn": 2048,
    "dimFullFrame": 2048 * 2048,
    "DEM_frequency": 10,  # [MHz]
    "FTI_science": 1000 / 15,  # [ms]
    "FTI_diagnostic": 240.0,  # [ms]
    "FTI_margin": 212.4,  # [ms]
    "overheadTime": 0.4644,  # [ms]
    "FOT_length": 20,
}
FULLFRAME_BYTES = 2 * DET_CONSTS["dimFullFrame"]

TSTAMP_MIN = 1561939200  # 2019-07-01T00:00:00+00:00
TSTAMP_TYPE = np.dtype([("tai_sec", int), ("sub_sec", int), ("dt", "O")])

module_logger = logging.getLogger("pyspex.lib.sci_tlm")


# - helper functions ------------------------
def subsec2musec(sub_sec: int) -> int:
    """Return subsec as microseconds."""
    return 100 * int(sub_sec / 65536 * 10000)


def mask2slice(mask: npt.NDArray[bool]) -> None | slice | tuple | npt.NDArray[bool]:
    """Try to slice (faster), instead of boolean indexing (slow)."""
    if np.all(~mask):
        return None
    if np.all(mask):
        return np.s_[:]  # read everything

    indx = np.nonzero(mask)[0]
    if np.all(np.diff(indx) == 1):
        # perform start-stop indexing
        return np.s_[indx[0] : indx[-1] + 1]

    # perform boolean indexing
    return mask


# - class SCItlm ----------------------------
class SCItlm:
    """Class to handle SPEXone Science-telemetry packets."""

    def __init__(self: SCItlm) -> None:
        """Initialize SCItlm object."""
        self.hdr: np.ndarray | None = None
        self.tlm: np.ndarray | None = None
        self.tstamp: np.ndarray | None = None
        self.images: tuple[np.ndarray, ...] | tuple[()] = ()

    def init_attrs(self: SCItlm) -> None:
        """Initialize class attributes."""
        self.hdr = None
        self.tlm = None
        self.tstamp = None
        self.images = ()

    @property
    def size(self: SCItlm) -> int:
        """Return number of elements."""
        return 0 if self.tlm is None else len(self.tlm)

    def copy(self: SCItlm) -> SCItlm:
        """Return deep-copy of SCItlm object."""
        sci = SCItlm()
        sci.hdr = self.hdr.copy()
        sci.tlm = self.tlm.copy()
        sci.tstamp = self.tstamp.copy()
        sci.images = copy(self.images)

    def sel(self: SCItlm, mask: np.NDArray[bool]) -> SCItlm:
        """Return subset of SCItlm object using a mask array."""
        sci = SCItlm()
        sci.hdr = self.hdr[mask]
        sci.tlm = self.tlm[mask]
        sci.tstamp = self.tstamp[mask]
        sci.images = tuple(x for x, y in zip(self.images, mask, strict=True) if y)
        return sci

    def extract_l0_sci(self: SCItlm, ccsds_sci: tuple, epoch: dt.datetime) -> int:
        """Extract SPEXone level-0 Science-telemetry data.

        Parameters
        ----------
        ccsds_sci :  tuple[np.ndarray, ...]
           SPEXone level-0 Science-telemetry packets
        epoch :  dt.datetime
           Epoch of the telemetry packets (1958 or 1970)

        Returns
        -------
        int
            number of detector frames

        """
        self.init_attrs()
        if not ccsds_sci:
            return 0

        n_frames = 0
        hdr_dtype = None
        hk_dtype = None
        found_start_first = False
        for buf in ccsds_sci:
            ccsds_hdr = CCSDShdr(buf["hdr"][0])
            if ccsds_hdr.grouping_flag == 1:
                found_start_first = True
                if n_frames == 0:
                    hdr_dtype = buf["hdr"].dtype
                    hk_dtype = buf["hk"].dtype
                    continue

            if not found_start_first:
                continue

            if ccsds_hdr.grouping_flag == 2:
                found_start_first = False
                n_frames += 1

        # do we have any complete detector images (Note ccsds_sci not empty!)?
        # print(f"n_frames: {n_frames}")
        if n_frames == 0:
            return 0

        # allocate memory
        self.hdr = np.empty(n_frames, dtype=hdr_dtype)
        self.tlm = np.empty(n_frames, dtype=hk_dtype)
        self.tstamp = np.empty(n_frames, dtype=TSTAMP_TYPE)

        # extract data from ccsds_sci
        ii = 0
        img = None
        found_start_first = False
        for buf in ccsds_sci:
            ccsds_hdr = CCSDShdr(buf["hdr"][0])
            if ccsds_hdr.grouping_flag == 1:
                found_start_first = True
                self.hdr[ii] = buf["hdr"]
                self.tlm[ii] = buf["hk"]
                self.tstamp[ii] = (
                    buf["icu_tm"]["tai_sec"][0],
                    buf["icu_tm"]["sub_sec"][0],
                    epoch
                    + dt.timedelta(
                        seconds=int(buf["icu_tm"]["tai_sec"][0]),
                        microseconds=subsec2musec(buf["icu_tm"]["sub_sec"][0]),
                    ),
                )
                img = (buf["frame"][0],)
                continue

            if not found_start_first:
                continue

            if ccsds_hdr.grouping_flag == 0:
                img += (buf["frame"][0],)
            elif ccsds_hdr.grouping_flag == 2:
                found_start_first = False
                img += (buf["frame"][0],)
                img = np.concatenate(img)
                if img.size == self.tlm[ii]["IMRLEN"] // 2:
                    self.images += (img,)
                    ii += 1
                else:
                    n_frames -= 1
                if ii == n_frames:
                    break

        # adjust number of frames for corrupted images
        if ii != self.hdr.size:
            self.hdr = self.hdr[:ii]
            self.tlm = self.tlm[:ii]
            self.tstamp = self.tstamp[:ii]

        return n_frames

    def extract_l1a_sci(self: SCItlm, fid: h5py.File, mps_id: int | None) -> None:
        """Extract data from SPEXone level-1a Science-telemetry packets.

        Parameters
        ----------
        fid :  h5py.File
           File pointer to a SPEXone level-1a product
        mps_id : int, optional
           Select data performed with MPS equals 'mps_id'

        """
        # pylint: disable=no-member
        self.init_attrs()

        # read science telemetry
        dset = fid["/science_data/detector_telemetry"]
        if mps_id is None:
            data_sel = np.s_[:]
        else:
            data_sel = mask2slice(dset.fields("MPS_ID")[:] == mps_id)
            if data_sel is None:
                return
        self.tlm = fid["/science_data/detector_telemetry"][data_sel]

        # determine time-stamps
        dset = fid["/image_attributes/icu_time_sec"]
        seconds = dset[data_sel]
        try:
            _ = dset.attrs["units"].index(b"1958")
        except ValueError:
            epoch = dt.datetime(1970, 1, 1, tzinfo=dt.UTC)
        else:
            epoch = dt.datetime(1958, 1, 1, tzinfo=dt.UTC)
            epoch -= dt.timedelta(seconds=get_leap_seconds(seconds[0]))
        subsec = fid["/image_attributes/icu_time_subsec"][data_sel]

        _dt = []
        for ii, sec in enumerate(seconds):
            msec_offs = self.readout_offset(ii)
            _dt.append(
                epoch
                + dt.timedelta(
                    seconds=int(sec),
                    milliseconds=-msec_offs,
                    microseconds=subsec2musec(subsec[ii]),
                )
            )

        self.tstamp = np.empty(len(seconds), dtype=TSTAMP_TYPE)
        self.tstamp["tai_sec"] = seconds
        self.tstamp["sub_sec"] = subsec
        self.tstamp["dt"] = _dt

        # read image data
        self.images = fid["/science_data/detector_images"][data_sel, :]

    def adc_gain(self: SCItlm, indx: int | None = None) -> np.ndarray:
        """Return ADC gain [Volt]."""
        if indx is None:
            indx = np.s_[:]
        return self.tlm["DET_ADCGAIN"][indx]

    def pga_gain(self: SCItlm, indx: int | None = None) -> np.ndarray:
        """Return PGA gain [Volt]."""
        if indx is None:
            indx = np.s_[:]

        # need first bit of address 121
        reg_pgagainfactor = self.tlm["DET_BLACKCOL"][indx] & 0x1

        reg_pgagain = self.tlm["DET_PGAGAIN"][indx]

        return (1 + 0.2 * reg_pgagain) * 2**reg_pgagainfactor

    def exposure_time(self: SCItlm, indx: int | None = None) -> np.ndarray:
        """Return exposure time [ms]."""
        if indx is None:
            indx = np.s_[:]
        return 129e-4 * (
            0.43 * self.tlm["DET_FOTLEN"][indx] + self.tlm["DET_EXPTIME"][indx]
        )

    def frame_period(self: SCItlm, indx: int) -> float:
        """Return frame period of detector measurement [ms]."""
        n_coad = self.tlm["REG_NCOADDFRAMES"][indx]
        # binning mode
        if self.tlm["REG_FULL_FRAME"][indx] == 2:
            return float(n_coad * DET_CONSTS["FTI_science"])

        # full-frame mode
        return float(
            n_coad
            * np.clip(
                DET_CONSTS["FTI_margin"]
                + DET_CONSTS["overheadTime"]
                + self.exposure_time(indx),
                a_min=DET_CONSTS["FTI_diagnostic"],
                a_max=None,
            )
        )

    def readout_offset(self: SCItlm, indx: int) -> float:
        """Return offset wrt start-of-integration [ms]."""
        n_coad = self.tlm["REG_NCOADDFRAMES"][indx]
        n_frm = (
            n_coad + 3
            if self.tlm["IMRLEN"][indx] == FULLFRAME_BYTES
            else 2 * n_coad + 2
        )

        return n_frm * self.frame_period(indx)

    def binning_table(self: SCItlm) -> np.ndarray:
        """Return binning table identifier (zero for full-frame images)."""
        bin_tbl = np.zeros(len(self.tlm), dtype="i1")
        _mm = (self.tlm["IMRLEN"] == FULLFRAME_BYTES) | (self.tlm["IMRLEN"] == 0)
        if np.sum(_mm) == len(self.tlm):
            return bin_tbl

        bin_tbl_start = self.tlm["REG_BINNING_TABLE_START"]
        bin_tbl[~_mm] = 1 + (bin_tbl_start[~_mm] - 0x80000000) // 0x400000
        return bin_tbl

    def digital_offset(self: SCItlm) -> np.ndarray:
        """Return digital offset including ADC offset [count]."""
        buff = self.tlm["DET_OFFSET"].astype("i4")
        buff[buff >= 8192] -= 16384

        return buff + 70

    def convert(self: SCItlm, key: str) -> np.ndarray:
        """Convert telemetry parameter to physical units.

        Parameters
        ----------
        key :  str
           Name of telemetry parameter

        Returns
        -------
        np.ndarray

        """
        parm = key.upper()
        if parm in ("HTR1_POWER", "HTR2_POWER", "HTR3_POWER", "HTR4_POWER"):
            parm = parm.replace("_POWER", "_I")
        else:
            if parm not in self.tlm.dtype.names:
                raise KeyError(
                    f"Parameter: {parm} not found" f" in {self.tlm.dtype.names}"
                )

        raw_data = np.array([x[parm] for x in self.tlm])
        return convert_hk(key.upper(), raw_data)



#
# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2019-2024 SRON - Netherlands Institute for Space Research
#    All Rights Reserved
#
# License:  BSD-3-Clause
"""Contains the class `DEMio` to read SPEXone CMV4000 detector data."""

from __future__ import annotations

__all__ = ["DEMio", "img_sec_of_day"]

from datetime import UTC, datetime, timedelta
from pathlib import Path
from typing import Any

import numpy as np

from .lib.tmtc_def import tmtc_dtype

# - global parameters ------------------------------
FULLFRAME_BYTES = 2 * 2048 * 2048


# - local functions --------------------------------
def det_dtype() -> np.dtype:
    """Return the registers of the SPEXone CMV4000 detector as a numpy dtype."""
    return np.dtype(
        [
            ("UNUSED_000", "u1"),
            ("NUMBER_LINES", "u1", (2,)),
            ("START1", "u1", (2,)),
            ("START2", "u1", (2,)),
            ("START3", "u1", (2,)),
            ("START4", "u1", (2,)),
            ("START5", "u1", (2,)),
            ("START6", "u1", (2,)),
            ("START7", "u1", (2,)),
            ("START8", "u1", (2,)),
            ("NUMBER_LINES1", "u1", (2,)),
            ("NUMBER_LINES2", "u1", (2,)),
            ("NUMBER_LINES3", "u1", (2,)),
            ("NUMBER_LINES4", "u1", (2,)),
            ("NUMBER_LINES5", "u1", (2,)),
            ("NUMBER_LINES6", "u1", (2,)),
            ("NUMBER_LINES7", "u1", (2,)),
            ("NUMBER_LINES8", "u1", (2,)),
            ("SUB_S", "u1", (2,)),
            ("SUB_A", "u1", (2,)),
            ("MONO", "u1"),  # 1 bits
            ("IMAGE_FLIPPING", "u1"),  # 2 bits
            ("INTE_SYNC", "u1"),  # 3 bits: Int_sync, Exp_dual, Exp_ext
            ("EXP_TIME", "u1", (3,)),
            ("EXP_STEP", "u1", (3,)),
            ("EXP_KP1", "u1", (3,)),
            ("EXP_KP2", "u1", (3,)),
            ("NR_SLOPES", "u1"),  # 2 bits
            ("EXP_SEQ", "u1"),
            ("EXP_TIME2", "u1", (3,)),
            ("EXP_STEP2", "u1", (3,)),
            ("UNUSED_062", "u1"),
            ("UNUSED_063", "u1"),
            ("UNUSED_064", "u1"),
            ("UNUSED_065", "u1"),
            ("UNUSED_066", "u1"),
            ("UNUSED_067", "u1"),
            ("UNUSED_068", "u1"),
            ("EXP2_SEQ", "u1"),
            ("NUMBER_FRAMES", "u1", (2,)),
            ("OUTPUT_MODE", "u1"),  # 2 bits
            ("FOT_LENGTH", "u1"),
            ("I_LVDS_REC", "u1"),  # 4 bits
            ("UNUSED_075", "u1"),
            ("UNUSED_076", "u1"),
            ("COL_CALIB", "u1"),  # 2 bits: Col_calib, ADC_calib
            ("TRAINING_PATTERN", "u1", (2,)),  # 12 bits
            ("CHANNEL_EN", "u1", (3,)),  # 19 bits
            ("I_LVDS", "u1"),  # 4 bits
            ("I_COL", "u1"),  # 4 bits
            ("I_COL_PRECH", "u1"),  # 4 bits
            ("I_ADC", "u1"),  # 4 bits
            ("I_AMP", "u1"),  # 4 bits
            ("VTF_L1", "u1"),  # 7 bits
            ("VLOW2", "u1"),  # 7 bits
            ("VLOW3", "u1"),  # 7 bits
            ("VRES_LOW", "u1"),  # 7 bits
            ("UNUSED_092", "u1"),
            ("UNUSED_093", "u1"),
            ("V_PRECH", "u1"),  # 7 bits
            ("V_REF", "u1"),  # 7 bits
            ("UNUSED_096", "u1"),
            ("UNUSED_097", "u1"),
            ("VRAMP1", "u1"),  # 7 bits
            ("VRAMP2", "u1"),  # 7 bits
            ("OFFSET", "u1", (2,)),  # 14 bits
            ("PGA_GAIN", "u1"),  # 2 bits
            ("ADC_GAIN", "u1"),
            ("UNUSED_104", "u1"),
            ("UNUSED_105", "u1"),
            ("UNUSED_106", "u1"),
            ("UNUSED_107", "u1"),
            ("T_DIG1", "u1"),  # 4 bits
            ("T_DIG2", "u1"),  # 4 bits
            ("UNUSED_110", "u1"),
            ("BIT_MODE", "u1"),  # 1 bits
            ("ADC_RESOLUTION", "u1"),  # 2 bits
            ("PLL_ENABLE", "u1"),  # 1 bits
            ("PLL_IN_FRE", "u1"),  # 2 bits
            ("PLL_BYPASS", "u1"),  # 1 bits
            ("PLL_RANGE", "u1"),  # 8 bits: PLL range(1), out_fre(3), div(4)
            ("PLL_LOAD", "u1"),
            ("DUMMY", "u1"),
            ("UNUSED_119", "u1"),
            ("UNUSED_120", "u1"),
            ("BLACK_COL_EN", "u1"),  # 2 bits: Black_col_en, PGA_gain
            ("UNUSED_122", "u1"),
            ("V_BLACKSUN", "u1"),  # 6 bits
            ("UNUSED_124", "u1"),
            ("UNUSED_125", "u1"),
            ("TEMP", "u1", (2,)),
        ]
    )


def img_sec_of_day(
    img_sec: np.ndarray, img_subsec: np.ndarray, img_hk: np.ndarray
) -> tuple[datetime, float | Any]:
    """Convert Image CCSDS timestamp to seconds after midnight.

    Parameters
    ----------
    img_sec : numpy array (dtype='u4')
        Seconds since 1970-01-01 (or 1958-01-01)
    img_subsec : numpy array (dtype='u2')
        Sub-seconds as (1 / 2**16) seconds
    img_hk :  numpy array
        DemHK telemetry packages

    Returns
    -------
    tuple
        reference day: datetime, sec_of_day: numpy.ndarray

    """
    # determine for the first timestamp the offset with last midnight [seconds]
    epoch = datetime(1970, 1, 1, tzinfo=UTC)
    tstamp0 = epoch + timedelta(seconds=int(img_sec[0]))
    ref_day = datetime(
        year=tstamp0.year, month=tstamp0.month, day=tstamp0.day, tzinfo=UTC
    )
    # seconds since midnight
    offs_sec = (ref_day - epoch).total_seconds()

    # Determine offset wrt start-of-integration (IMRO + 1)
    # Where the default is defined as IMRO:
    #  [full-frame] COADDD + 2  (no typo, this is valid for the later MPS's)
    #  [binned] 2 * COADD + 1   (always valid)
    offs_msec = 0
    if img_hk["ICUSWVER"][0] > 0x123:
        imro = np.empty(img_hk.size, dtype=float)
        _mm = img_hk["IMRLEN"] == FULLFRAME_BYTES
        imro[_mm] = img_hk["REG_NCOADDFRAMES"][_mm] + 2
        imro[~_mm] = 2 * img_hk["REG_NCOADDFRAMES"][~_mm] + 1
        offs_msec = img_hk["FTI"] * (imro + 1) / 10

    # return seconds since midnight
    return ref_day, img_sec - offs_sec + img_subsec / 65536 - offs_msec / 1000


# - class DEMio -------------------------
class DEMio:
    """Read SPEXone DEM output.

    Parameters
    ----------
    flname :  str
        filename with header or binary data of DEM measurement

    Examples
    --------
    Read data from a SPEXone DEM binary file::

    > dem = DEMio(dem_file)
    > img_hk = dem.get_sci_hk()
    > img_data = dem.get_data()

    """

    def __init__(self: DEMio, flname: str) -> None:
        """Initialize DEMio object."""
        self.__hdr = None
        if flname.endswith("a.txt"):
            self.bin_file = flname.replace("a.txt", "b.bin")
            self.hdr_file = flname
        elif flname.endswith("b.bin"):
            self.bin_file = flname
            self.hdr_file = flname.replace("b.bin", "a.txt")
        else:
            raise RuntimeError(f"invalid filename: {flname}")

        if Path(self.hdr_file).is_file():
            self.__get_hdr()

    def __get_hdr(self: DEMio) -> None:
        """Read DEM header data."""
        self.__hdr = np.zeros((1,), dtype=det_dtype())
        with open(self.hdr_file, encoding="ascii", errors="ignore") as fp:
            for line in fp:
                columns = line[:-1].split(",")
                if columns[0] == "Reg":
                    continue

                # Fix possible errors in Name
                name = columns[2].replace(" [", "[")
                name = name.replace("_ ", "_").replace(" ", "_")

                value = int(columns[-1])
                indx = -1
                if (
                    columns[2].endswith(":0]")
                    or columns[2].endswith("[0]")
                    or columns[2].endswith("[2]")
                ):
                    name = name.split("[")[0]
                    indx = 0
                elif columns[2].endswith(":8]"):
                    name = name.split("[")[0]
                    indx = 1
                elif columns[2].endswith(":16]"):
                    name = name.split("[")[0]
                    indx = 2
                elif name == "Unused":
                    if columns[0] == "86":
                        continue
                    name = f"Unused_{int(columns[0]):03d}"

                key = name.upper()
                if isinstance(self.__hdr[0][key], np.ndarray):
                    self.__hdr[0][key][indx] = value
                else:
                    self.__hdr[0][key] = value

    @property
    def hdr(self: DEMio) -> np.ndarray | None:
        """Return DEM header as numpy compound array."""
        if self.__hdr is None:
            return None

        return self.__hdr[0]

    @property
    def number_lines(self: DEMio) -> int:
        """Return number of lines (rows).

        Register address: [1, 2]
        """
        return self.hdr["NUMBER_LINES"][0] + (self.hdr["NUMBER_LINES"][1] << 8)

    @property
    def number_channels(self: DEMio) -> int:
        """Return number of LVDS channels used."""
        return 2 ** (4 - (self.hdr["OUTPUT_MODE"] & 0x3))

    @property
    def lvds_clock(self: DEMio) -> bool:
        """Return flag for LVDS clock (0: disable, 1: enable).

        Register address: 82
        """
        return (
            (self.hdr["PLL_ENABLE"] & 0x3) == 0
            and (self.hdr["PLL_BYPASS"] & 0x3) != 0
            and (self.hdr["CHANNEL_EN"][2] & 0x4) != 0
        )

    def pll_control(self: DEMio) -> tuple:
        """Return PLL control parameters: pll_range, pll_out_fre, pll_div.

        PLL_range:    range (0 or 1)
        PLL_out_fre:  output frequency (0, 1, 2 or 5)
        PLL_div:      9 (10 bit) or 11 (12 bit)

        Register address: 116
        """
        pll_div = self.hdr["PLL_RANGE"] & 0xF  # bit [0:4]
        pll_out_fre = (self.hdr["PLL_RANGE"] >> 4) & 0x7  # bit [4:7]
        pll_range = self.hdr["PLL_RANGE"] >> 7  # bit [7]

        return pll_range, pll_out_fre, pll_div

    @property
    def exp_control(self: DEMio) -> tuple:
        """Exposure time control parameters: (inte_sync, exp_dual, exp_ext).

        Register address: 41
        """
        inte_sync = (self.hdr["INTE_SYNC"] >> 2) & 0x1
        exp_dual = (self.hdr["INTE_SYNC"] >> 1) & 0x1
        exp_ext = self.hdr["INTE_SYNC"] & 0x1

        return inte_sync, exp_dual, exp_ext

    @property
    def offset(self: DEMio) -> int:
        """Return digital offset including ADC offset.

        Register address: [100, 101]
        """
        val = (self.hdr["OFFSET"][1] << 8) + self.hdr["OFFSET"][0]

        return 70 + (val if val < 8192 else val - 16384)

    @property
    def pga_gain(self: DEMio) -> float:
        """Return PGA gain (Volt).

        Register address: 102
        """
        reg_pgagain = self.hdr["PGA_GAIN"]
        # need first bit of address 121
        reg_pgagainfactor = self.hdr["BLACK_COL_EN"] & 0x1

        return (1 + 0.2 * reg_pgagain) * 2**reg_pgagainfactor

    @property
    def temp_detector(self: DEMio) -> int:
        """Return detector temperature as raw counts.

        Notes
        -----
        Uncalibrated conversion: ((1184 - 1066) * 0.3 * 40 / 40Mhz) + offs [K]

        """
        return (self.hdr["TEMP"][1] << 8) + self.hdr["TEMP"][0]

    def exp_time(self: DEMio, t_mcp: float = 1e-7) -> float:
        """Return pixel exposure time [s]."""
        # Nominal fot_length = 20, except for very short exposure_time
        reg_fot = self.hdr["FOT_LENGTH"]

        reg_exptime = (
            (self.hdr["EXP_TIME"][2] << 16)
            + (self.hdr["EXP_TIME"][1] << 8)
            + self.hdr["EXP_TIME"][0]
        )

        return 129 * t_mcp * (0.43 * reg_fot + reg_exptime)

    def fot_time(self: DEMio, t_mcp: float = 1e-7) -> float:
        """Return frame overhead time [s]."""
        # Nominal fot_length = 20, except for very short exposure_time
        reg_fot = self.hdr["FOT_LENGTH"]

        return 129 * t_mcp * (reg_fot + 2 * (16 // self.number_channels))

    def rot_time(self: DEMio, t_mcp: float = 1e-7) -> float:
        """Return image read-out time [s]."""
        return 129 * t_mcp * (16 // self.number_channels) * self.number_lines

    def frame_period(self: DEMio, n_coad: int = 1) -> float:
        """Return frame period [s]."""
        return 2.38 + (n_coad * (self.exp_time() + self.fot_time() + self.rot_time()))

    def get_sci_hk(self: DEMio) -> np.ndarray:
        """Return Science telemetry.

        A subset of MPS and housekeeping parameters.

        Returns
        -------
        numpy.ndarray

        """

        def convert_val(kk: str) -> int:
            """Convert byte array to integer."""
            val = 0
            for ii, bval in enumerate(self.__hdr[0][kk]):
                val += bval << (ii * 8)

            return val

        # convert original detector parameter values to telemetry parameters
        convert_det_params = {
            "DET_NUMLINES": convert_val("NUMBER_LINES"),
            "DET_START1": convert_val("START1"),
            "DET_START2": convert_val("START2"),
            "DET_START3": convert_val("START3"),
            "DET_START4": convert_val("START4"),
            "DET_START5": convert_val("START5"),
            "DET_START6": convert_val("START6"),
            "DET_START7": convert_val("START7"),
            "DET_START8": convert_val("START8"),
            "DET_NUMLINES1": convert_val("NUMBER_LINES1"),
            "DET_NUMLINES2": convert_val("NUMBER_LINES2"),
            "DET_NUMLINES3": convert_val("NUMBER_LINES3"),
            "DET_NUMLINES4": convert_val("NUMBER_LINES4"),
            "DET_NUMLINES5": convert_val("NUMBER_LINES5"),
            "DET_NUMLINES6": convert_val("NUMBER_LINES6"),
            "DET_NUMLINES7": convert_val("NUMBER_LINES7"),
            "DET_NUMLINES8": convert_val("NUMBER_LINES8"),
            "DET_SUBS": convert_val("SUB_S"),
            "DET_SUBA": convert_val("SUB_A"),
            "DET_MONO": self.__hdr[0]["MONO"],
            "DET_IMFLIP": self.__hdr[0]["IMAGE_FLIPPING"],
            "DET_EXPCNTR": self.__hdr[0]["INTE_SYNC"],
            "DET_EXPTIME": convert_val("EXP_TIME"),
            "DET_EXPSTEP": convert_val("EXP_STEP"),
            "DET_KP1": convert_val("EXP_KP1"),
            "DET_KP2": convert_val("EXP_KP2"),
            "DET_NOFSLOPES": self.__hdr[0]["NR_SLOPES"],
            "DET_EXPSEQ": self.__hdr[0]["EXP_SEQ"],
            "DET_EXPTIME2": convert_val("EXP_TIME2"),
            "DET_EXPSTEP2": convert_val("EXP_STEP2"),
            "DET_EXP2_SEQ": self.__hdr[0]["EXP2_SEQ"],
            "DET_NOFFRAMES": convert_val("NUMBER_FRAMES"),
            "DET_OUTMODE": self.__hdr[0]["OUTPUT_MODE"],
            "DET_FOTLEN": self.__hdr[0]["FOT_LENGTH"],
            "DET_ILVDSRCVR": self.__hdr[0]["I_LVDS_REC"],
            "DET_CALIB": self.__hdr[0]["COL_CALIB"],
            "DET_TRAINPTRN": convert_val("TRAINING_PATTERN"),
            "DET_CHENA": convert_val("CHANNEL_EN"),
            "DET_ILVDS": self.__hdr[0]["I_LVDS"],
            "DET_ICOL": self.__hdr[0]["I_COL"],
            "DET_ICOLPR": self.__hdr[0]["I_COL_PRECH"],
            "DET_IADC": self.__hdr[0]["I_ADC"],
            "DET_IAMP": self.__hdr[0]["I_AMP"],
            "DET_VTFL1": self.__hdr[0]["VTF_L1"],
            "DET_VTFL2": self.__hdr[0]["VLOW2"],
            "DET_VTFL3": self.__hdr[0]["VLOW3"],
            "DET_VRSTL": self.__hdr[0]["VRES_LOW"],
            "DET_VPRECH": self.__hdr[0]["V_PRECH"],
            "DET_VREF": self.__hdr[0]["V_REF"],
            "DET_VRAMP1": self.__hdr[0]["VRAMP1"],
            "DET_VRAMP2": self.__hdr[0]["VRAMP2"],
            "DET_OFFSET": convert_val("OFFSET"),
            "DET_PGAGAIN": self.__hdr[0]["PGA_GAIN"],
            "DET_ADCGAIN": self.__hdr[0]["ADC_GAIN"],
            "DET_TDIG1": self.__hdr[0]["T_DIG1"],
            "DET_TDIG2": self.__hdr[0]["T_DIG2"],
            "DET_BITMODE": self.__hdr[0]["BIT_MODE"],
            "DET_ADCRES": self.__hdr[0]["ADC_RESOLUTION"],
            "DET_PLLENA": self.__hdr[0]["PLL_ENABLE"],
            "DET_PLLINFRE": self.__hdr[0]["PLL_IN_FRE"],
            "DET_PLLBYP": self.__hdr[0]["PLL_BYPASS"],
            "DET_PLLRATE": self.__hdr[0]["PLL_RANGE"],
            "DET_PLLLOAD": self.__hdr[0]["PLL_LOAD"],
            "DET_DETDUM": self.__hdr[0]["DUMMY"],
            "DET_BLACKCOL": self.__hdr[0]["BLACK_COL_EN"],
            "DET_VBLACKSUN": self.__hdr[0]["V_BLACKSUN"],
            "DET_T": convert_val("TEMP"),
        }

        sci_hk = np.zeros((1,), dtype=tmtc_dtype(0x350))
        sci_hk[0]["REG_FULL_FRAME"] = 1
        sci_hk[0]["REG_CMV_OUTPUTMODE"] = 3
        for key, value in convert_det_params.items():
            sci_hk[0][key] = value

        return sci_hk

    def get_data(self: DEMio, numlines: int = None) -> np.ndarray:
        """Return data of a detector frame (numpy uint16 array).

        Parameters
        ----------
        numlines : int, optional
           Provide number of detector rows when no headerfile is present

        Returns
        -------
        numpy.ndarray:
           data of a detector frame, dtype np.uint16

        """
        if numlines is None:
            # obtain number of rows
            numlines = self.number_lines

        # Read binary big-endian data
        return np.fromfile(self.bin_file, dtype=">u2").reshape(numlines, -1)



#
# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2019-2024 SRON - Netherlands Institute for Space Research
#    All Rights Reserved
#
# License:  BSD-3-Clause
"""Defines the spectral dependent DolP of the Moxtek polarizer."""

from __future__ import annotations

__all__ = ["gsfc_polarizer"]

import numpy as np
import xarray as xr

# - global parameters ------------------------------
DOLP_ATTRS = {
    "instrument": "Moxtek polarizer characterisation",
    "comment": "neglect when attribute DolP=0",
}

WAVE = [360, 380, 410, 440, 460, 488, 508.5, 532, 580, 770, 870]
DOLP = [
    0.9805,
    0.9833,
    0.9875,
    0.9913,
    0.9932,
    0.9948,
    0.9962,
    0.9972,
    0.9972,
    0.9983,
    0.9972,
]
RATIO = [101.6, 118.9, 159.0, 229.5, 292.6, 383.6, 530.6, 713.3, 710.1, 1193.0, 723.0]


# - local functions ----------------------------
def gsfc_polarizer() -> xr.Dataset:
    """Define table for spectral dependent DolP."""
    xar_wv = xr.DataArray(
        np.array(WAVE, dtype="f4"),
        coords={"wavelength": WAVE},
        attrs={
            "longname": "wavelength grid",
            "units": "nm",
            "comment": "wavelength annotation",
        },
    )
    xar_dolp = xr.DataArray(
        np.array(DOLP, dtype="f4"),
        coords={"wavelength": WAVE},
        attrs={"longname": "DoLP", "units": "1"},
    )
    xar_ratio = xr.DataArray(
        np.array(RATIO, dtype="f4"),
        coords={"wavelength": WAVE},
        attrs={"longname": "Extinction ratio", "units": "1"},
    )

    return xr.Dataset(
        {"wavelength": xar_wv, "DoLP": xar_dolp, "Extinction_Ratio": xar_ratio},
        attrs=DOLP_ATTRS,
    )


def __test(l1a_file: str) -> None:
    """Small function to test this module."""
    # Create a netCDF4 file containing the spectralDoLP
    xds = gsfc_polarizer()
    xds.to_netcdf(l1a_file, mode="w", format="NETCDF4", group="/gse_data/Polarizer")


# --------------------------------------------------
if __name__ == "__main__":
    print("---------- SHOW DATASET ----------")
    print(gsfc_polarizer())
    print("---------- WRITE DATASET ----------")
    __test("test_netcdf.nc")



# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2019-2024 SRON - Netherlands Institute for Space Research
#   All Rights Reserved
#
# License:  BSD-3-Clause
"""Necessary but empty file."""



#
# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2019-2024 SRON - Netherlands Institute for Space Research
#    All Rights Reserved
#
# License:  BSD-3-Clause
"""Provides the common global-attributes for SPEXone Level-1 products."""

from __future__ import annotations

__all__ = ["attrs_def"]

import datetime as dt

from . import pyspex_version


# - main functions --------------------------------
def attrs_def(inflight: bool = True, origin: str | None = None) -> dict:
    """Define global attributes of a SPEXone Level-1A product.

    Parameters
    ----------
    inflight : bool, default=True
       Flag for in-flight or on-ground products
    origin : str
       Product origin: 'SRON' or 'NASA'

    Returns
    -------
    dict
       Global attributes for a Level-1A product

    """
    if origin is None:
        origin = "NASA" if inflight else "SRON"

    res = {
        "title": "PACE SPEXone Level-1A data",
        "platform": "PACE",
        "instrument": "SPEXone",
        "institution": (
            "NASA Goddard Space Flight Center," " Ocean Biology Processing Group"
        ),
        "license": (
            "http://science.nasa.gov/earth-science/"
            "earth-science-data/data-information-policy/"
        ),
        "naming_authority": "gov.nasa.gsfc.sci.oceancolor",
        "keyword_vocabulary": (
            "NASA Global Change Master Directory (GCMD)" " Science Keywords"
        ),
        "stdname_vocabulary": (
            "NetCDF Climate and Forecast (CF)" " Metadata Convention"
        ),
        "standard_name_vocabulary": "CF Standard Name Table v79",
        "conventions": "CF-1.8 ACDD-1.3",
        "identifier_product_doi_authority": "http://dx.doi.org/",
        "identifier_product_doi": "https://doi.org/10.5281/zenodo.5705691",
        "creator_name": "NASA/GSFC",
        "creator_email": "data@oceancolor.gsfc.nasa.gov",
        "creator_url": "http://oceancolor.gsfc.nasa.gov",
        "project": "PACE Project",
        "publisher_name": "NASA/GSFC",
        "publisher_email": "data@oceancolor.gsfc.nasa.gov",
        "publisher_url": "http://oceancolor.gsfc.nasa.gov",
        "cdm_data_type": ("One orbit swath or granule" if inflight else "granule"),
        "cdl_version_date": "2021-09-10",
        "product_name": None,
        "processing_level": "L1A",
        "processing_version": "",
        "date_created": dt.datetime.now(dt.UTC).isoformat(timespec="milliseconds"),
        "software_name": "SPEXone L0-L1A processor",
        "software_url": "https://github.com/rmvanhees/pyspex",
        "software_version": pyspex_version(),
        "history": "spx1_level01a.py",
        "start_direction": "Ascending" if inflight else None,
        "end_direction": "Ascending" if inflight else None,
        "time_coverage_start": "yyyy-mm-ddTHH:MM:DD",
        "time_coverage_end": "yyyy-mm-ddTHH:MM:DD",
    }

    if origin == "SRON":
        res["title"] = "SPEXone Level-1A data"
        res["institution"] = "SRON Netherlands Institute for Space Research"
        res["creator_name"] = "SRON/Earth"
        res["creator_email"] = "SPEXone-MPC@sron.nl"
        res["creator_url"] = "https://www.sron.nl/missions-earth/pace-spexone"
        res["publisher_name"] = "SRON/Earth"
        res["publisher_email"] = "SPEXone-MPC@sron.nl"
        res["publisher_url"] = "https://www.sron.nl/missions-earth/pace-spexone"

    return res



#
# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2019-2024 SRON - Netherlands Institute for Space Research
#    All Rights Reserved
#
# License:  BSD-3-Clause
"""Collect OGSE data in a database.

This module contain routines to read reference diode measurements and
wavelength monitor data. These data are supposed to be written to a HDF5
database. From these database selected data can be added to a SPEXone Level-1a
product.
"""

from __future__ import annotations

__all__ = ["read_ref_diode", "read_wav_mon", "add_ogse_ref_diode", "add_ogse_wav_mon"]

from datetime import datetime
from io import StringIO
from pathlib import Path
from typing import Any

import h5py
import numpy as np
from xarray import DataArray, Dataset, open_dataset

# - global parameters ------------------------------


# ---------- CREATE OGSE DATABASES ----------
def read_ref_diode(ogse_dir: Path, file_list: list, verbose: bool = False) -> Dataset:
    """Read reference diode data into a xarray.Dataset.

    (input: comma separated values).
    """
    data = None
    unit_dict = {
        "Unix Time (s)": ("seconds", "f8"),
        "Excel Time (d)": ("days", "f4"),
        "Normalized Time (s)": ("seconds", "f8"),
        "amps": ("A", "f4"),
        "scaled": ("1", "f4"),
        "last_zero": ("seconds", "f8"),
        "lamp_on": ("flag", "b"),
        "voltage": ("V", "f4"),
        "current": ("A", "f4"),
    }

    fields_to_skip = (
        "",
        "averaging",
        "lamp_on_counter",
        "record_timestamp",
        "scale",
        "tempC",
    )

    names = []
    units = []
    for flname in file_list:
        formats = []
        usecols = ()
        all_valid_lines = ""
        with open(ogse_dir / flname, encoding="ascii") as fid:
            while True:
                line = fid.readline().strip()
                if not line.startswith("Unix"):
                    continue

                fields = line.split(",")
                for ii, field in enumerate(fields):
                    if field in fields_to_skip:
                        continue
                    res = field.strip().split(" (")
                    res[0] = res[0].replace(" ", "_")
                    if res[0] in names:
                        continue
                    names.append(res[0])
                    units.append(unit_dict[field][0])
                    formats.append(unit_dict[field][1])
                    usecols += (ii,)
                break

            if verbose:
                print(len(names), names)
                print(len(units), units)
                print(len(formats), formats)
                print(len(usecols), usecols)
                # return (data, units)

            while True:
                line = fid.readline()
                if not line:
                    break
                if ",,,," not in line:
                    all_valid_lines += line.replace(",,", ",0,")

        # read data (skip first line)
        res = np.loadtxt(
            StringIO(all_valid_lines),
            delimiter=",",
            skiprows=1,
            usecols=usecols,
            converters={0: lambda s: float(s) - 3600},
            dtype={"names": names, "formats": formats},
        )
        data = res if data is None else np.concatenate((data, res))

    if verbose:
        print("data :", len(data))

    time_key = "Unix_Time"
    res = {
        "time": DataArray(
            data[time_key],
            coords={"time": data[time_key]},
            attrs={"longname": "time", "units": "seconds since 1970-01-01 00:00:00"},
        )
    }
    for ii, key in enumerate(names):
        if key == time_key:
            continue

        res[key] = DataArray(
            data[key],
            coords={"time": data[time_key]},
            attrs={"longname": key, "units": units[ii]},
        )

    xds = Dataset(res).sortby("time")
    xds.attrs = {
        "source": "Calibrated diode inside integrated sphere",
        "comment": "Generated on SRON clean-room tablet",
    }
    return xds


# ---------------
def read_wav_mon(ogse_dir: Path, file_list: list, verbose: bool = False) -> Dataset:
    """Read wavelength monitor data into a xarray.Dataset.

    (input comma separated values).
    """

    def byte_to_timestamp(str_date: bytes) -> float:
        """Return a timestamp (UTC) from a byte-string.

        Parameters
        ----------
        str_date : byte
            date-time string in ISO format with timezone CET

        Returns
        -------
        timestamp :  float
            Unix timestamp in UTC

        """
        buff = str_date.strip().decode("ascii") + "+01:00"
        return datetime.strptime(buff, "%Y-%m-%dT%H:%M:%S.%f%z").timestamp()

    names = ("timestamp", "spectrum")
    formats = None
    data = None
    wavelength = None
    t_intg = None
    n_avg = None

    for flname in file_list:
        with open(ogse_dir / flname, encoding="ascii") as fid:
            while True:
                line = fid.readline().strip()

                if line.startswith("Integration"):
                    fields = line.split(":")
                    scalar_t_intg = float(fields[1])
                    continue

                if line.startswith("Averaging"):
                    fields = line.split(":")
                    scalar_n_avg = int(fields[1])
                    continue

                if line.startswith("Timestamp"):
                    fields = line.split(",")
                    wavelength = np.array(fields[1:], dtype=float)

                    # define dtype of the data
                    formats = ("f8", f"{len(wavelength)}f8")
                    break

            if verbose:
                print(f"Integration time: {t_intg}ms")
                print(f"Averaging Nr.: {n_avg}")
                print(len(names), names)
                print(len(formats), formats)
                # return (None, None)

            res = np.loadtxt(
                fid,
                delimiter=",",
                converters={0: byte_to_timestamp},
                dtype={"names": names, "formats": formats},
            )

        if verbose:
            print("data :", len(res))

        if data is None:
            data = res
            t_intg = np.full(res.size, scalar_t_intg)
            n_avg = np.full(res.size, scalar_n_avg)
        else:
            data = np.concatenate((data, res))
            t_intg = np.concatenate((t_intg, np.full(res.size, scalar_t_intg)))
            n_avg = np.concatenate((n_avg, np.full(res.size, scalar_n_avg)))

    time_key = "timestamp"
    res = {
        "time": DataArray(
            data[time_key],
            coords={"time": data[time_key]},
            attrs={"longname": "time", "units": "seconds since 1970-01-01 00:00:00"},
        ),
        "wavelength": DataArray(
            wavelength,
            coords={"wavelength": wavelength},
            attrs={"longname": "wavelength grid", "units": "nm"},
        ),
        "t_intg": DataArray(
            t_intg.astype("i2"),
            dims=["time"],
            attrs={"longname": "Integration time", "units": "nm"},
        ),
        "n_avg": DataArray(
            n_avg.astype("i2"),
            dims=["time"],
            attrs={"longname": "Averaging number", "units": "1"},
        ),
        "spectrum": DataArray(
            data["spectrum"],
            dims=["time", "wavelength"],
            attrs={"longname": "radiance spectrum", "units": "W/(m^2.sr.nm)"},
        ),
    }

    xds = Dataset(res).sortby("time")
    xds.attrs = {
        "source": "Avantes fibre spectrometer",
        "comment": "Generated on SRON clean-room tablet",
    }
    return xds


# ----- SELECT OGSE DATA FROM DATABASE AND ADD TO L1A PRODUCT -----
def read_date_stats() -> tuple:
    """Read output of program 'date' executed at freckle (ITOS) and shogun (SRON)."""
    if Path("/array/slot1F/spex_one/OCAL/date_stats").is_dir():
        data_dir = Path("/array/slot2B/spex_ocal/ambient/date_stats")
    elif Path("./date_stats").is_dir():
        data_dir = Path("./date_stats")
    else:
        data_dir = Path("/nfs/SPEXone/ocal/ambient/date_stats")
    flname = data_dir / "cmp_date_egse_itos2.txt"

    all_valid_lines = ""
    with open(flname, encoding="ascii") as fid:
        names = fid.readline().strip().lstrip(" #").split("\t\t\t")
        formats = len(names) * ("f8",)

        while True:
            line = fid.readline()
            if not line:
                break
            fields = line.split("\t")
            if "" not in fields:
                all_valid_lines += line

    res = np.loadtxt(
        StringIO(all_valid_lines), dtype={"names": names, "formats": formats}
    )

    # a constant difference between the SRON/ITOS timestamps is introduced
    # by the calls of 'date' via 'ssh'. This difference is estimated at 350ms.
    t_itos = (1000 * res["ITOS"] - 350).astype(int)
    t_sron = (1000 * res["SRON(1)"]).astype(int)

    return t_itos.astype("datetime64[ms]"), t_sron.astype("datetime64[ms]")


def clock_offset(l1a_file: Path) -> tuple[Any, int | Any, Any]:
    """Derive offset between msmt_start/msmt_stop and the SRON clock."""
    # determine duration of the measurement (ITOS clock)
    with h5py.File(l1a_file, "r") as fid:
        # pylint: disable=unsubscriptable-object
        res = fid.attrs["input_files"]
        if isinstance(res, bytes):
            input_file = Path(res.decode("ascii")).stem.rstrip("_hk")
        else:
            input_file = Path(res[0]).stem.rstrip("_hk")
        # pylint: disable=no-member
        msmt_start = np.datetime64(
            fid.attrs["time_coverage_start"].decode("ascii").split("+")[0]
        )
        msmt_stop = np.datetime64(
            fid.attrs["time_coverage_end"].decode("ascii").split("+")[0]
        )

    duration = (msmt_stop - msmt_start).astype("timedelta64[s]") + 1
    # print('duration: ', duration)

    # use the timestamp in the filename to correct ICU time
    date_start = datetime.strptime(input_file.split("_")[-1], "%Y%m%dT%H%M%S.%f")
    msmt_start = np.datetime64(date_start.isoformat()).astype("datetime64[s]")
    msmt_stop = msmt_start + duration
    # print('msmt: ', msmt_start, msmt_stop)

    # correct msmt_start and msmt_stop to SRON clock
    t_itos, t_sron = read_date_stats()
    indx = np.argsort(np.abs(t_itos - t_sron))[0]
    t_diff = t_sron[indx] - t_itos[indx]
    # print('T_sron - T_itos:', t_diff)

    return msmt_start, msmt_stop, t_diff


def add_ogse_ref_diode(ref_db: Path, l1a_file: Path) -> None:
    """Select reference data taken during a measurement and add to a L1A product."""
    # msmt_start and msmt_stop are generated with the ITOS clock
    msmt_start, msmt_stop, t_diff = clock_offset(l1a_file)

    # ref_time is generated with the SRON clock
    xds = open_dataset(ref_db, group="/gse_data/ReferenceDiode")
    ref_time = xds["time"].values - t_diff
    indx = np.where((ref_time >= msmt_start) & (ref_time <= msmt_stop))[0]
    if indx.size == 0:
        print("ReferenceDiode", ref_time.min(), msmt_start, msmt_stop, ref_time.max())
        print("[WARNING]: no reference-diode data found")
        return

    # update Level-1A product with OGSE/EGSE information
    xds = xds.isel(time=indx)
    xds.to_netcdf(l1a_file, mode="a", group="/gse_data/ReferenceDiode")


def add_ogse_wav_mon(ref_db: Path, l1a_file: Path) -> None:
    """Select reference data taken during a measurement and add to a L1A product."""
    # msmt_start and msmt_stop are generated with the ITOS clock
    msmt_start, msmt_stop, t_diff = clock_offset(l1a_file)

    # ref_time is generated with the SRON clock
    xds = open_dataset(ref_db, group="/gse_data/WaveMonitor")
    ref_time = xds["time"].values - t_diff
    mask = (ref_time >= msmt_start) & (ref_time <= msmt_stop)
    if mask.sum() == 0:
        print("WaveMonitor", ref_time.min(), msmt_start, msmt_stop, ref_time.max())
        print("[WARNING]: no wavelength monitoring data found")
        return

    # update Level-1A product with OGSE/EGSE information
    xds = xds.isel(time=mask.nonzero()[0])
    xds.to_netcdf(l1a_file, mode="a", group="/gse_data/WaveMonitor")


def __test() -> None:
    """Small function to test this module."""
    ogse_dir = Path(
        "/data/richardh/SPEXone/ambient/polarimetric/calibration/light_level/Logs/"
    )

    file_list = ["POLARIMETRIC_LIGHT_LEVELS_ref_det_session174.csv"]
    print("---------- SHOW DATASET ----------")
    print(read_ref_diode(ogse_dir, file_list, verbose=True))
    # print('---------- WRITE DATASET ----------')
    # create_ref_diode_db(ogse_dir, file_list)

    file_list = ["POLARIMETRIC_LIGHTLEVELS_3_Avantes_20210211T165045.csv"]
    print("---------- SHOW DATASET ----------")
    print(read_wav_mon(ogse_dir, file_list, verbose=True))
    # print('---------- WRITE DATASET ----------')
    # create_wav_mon_db(ogse_dir, file_list)


# --------------------------------------------------
if __name__ == "__main__":
    __test()



#
# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2019-2024 SRON - Netherlands Institute for Space Research
#    All Rights Reserved
#
# License:  BSD-3-Clause
"""Read Primary and Secondary CCSDS headers and obtain their parameters."""

from __future__ import annotations

__all__ = ["CCSDShdr"]

import datetime as dt

import numpy as np

from .tmtc_def import tmtc_dtype


class CCSDShdr:
    """Read CCSDS telemetry packet structure.

    Which consists of the primary header: version, type, apid, grouping flag,
    sequence count and packet length, and the secondary header: tai_sec and
    sub_sec.

    """

    def __init__(self: CCSDShdr, hdr: np.ndarray | None = None) -> None:
        """Initialise the class instance.

        Parameters
        ----------
        hdr :  np.ndarray, optional
           CCSDS primary and secondary headers

        """
        self.__dtype = None
        self.__hdr = None
        if hdr is not None:
            self.__hdr = hdr
            self.__dtype = hdr.dtype

    def _tm_raw_(self: CCSDShdr) -> np.dtype:  # ApID unknown
        """Return data-type of unknown packet, just a header and byte data."""
        return np.dtype(
            [("hdr", self.__hdr.dtype), ("Data", "u1", (self.__hdr["length"] - 5))]
        )

    def _tm_800_(self: CCSDShdr) -> np.dtype:  # ApID = 0x320
        """Return data-type of NomHk packet."""
        return np.dtype([("hdr", self.__hdr.dtype), ("hk", tmtc_dtype(0x320))])

    def _tm_802_(self: CCSDShdr) -> np.dtype:  # ApID = 0x322
        """Return data-type of DemHk packet."""
        return np.dtype([("hdr", self.__hdr.dtype), ("hk", tmtc_dtype(0x322))])

    def _tm_817_(self: CCSDShdr) -> np.dtype:  # ApID = 0x331
        """Return data-type of TcAccept packet."""
        return np.dtype(
            [("hdr", self.__hdr.dtype), ("TcPacketId", ">u2"), ("TcSeqControl", ">u2")]
        )

    def _tm_818_(self: CCSDShdr) -> np.dtype:  # ApID = 0x332
        """Return data-type of TcReject packet."""
        return np.dtype(
            [
                ("hdr", self.__hdr.dtype),
                ("TcPacketId", ">u2"),
                ("TcSeqControl", ">u2"),
                ("TcRejectCode", ">u2"),
                ("RejectParameter1", ">u2"),
                ("RejectParameter2", ">u2"),
            ]
        )

    def _tm_819_(self: CCSDShdr) -> np.dtype:  # ApID = 0x333
        """Return data-type of TcExecute packet."""
        return np.dtype(
            [("hdr", self.__hdr.dtype), ("TcPacketId", ">u2"), ("TcSeqControl", ">u2")]
        )

    def _tm_820_(self: CCSDShdr) -> np.dtype:  # ApID = 0x334
        """Return data-type of TcFail packet."""
        return np.dtype(
            [
                ("hdr", self.__hdr.dtype),
                ("TcPacketId", ">u2"),
                ("TcSeqControl", ">u2"),
                ("TcFailCode", ">u2"),
                ("FailParameter1", ">u2"),
                ("FailParameter2", ">u2"),
            ]
        )

    def _tm_821_(self: CCSDShdr) -> np.dtype:  # ApID = 0x335
        """Return data-type of EventRp packet."""
        return np.dtype(
            [
                ("hdr", self.__hdr.dtype),
                ("Event_ID", "u1"),
                ("Event_Sev", "u1"),
                ("Word1", ">u2"),
                ("Word2", ">u2"),
                ("Word3", ">u2"),
                ("Word4", ">u2"),
                ("Word5", ">u2"),
                ("Word6", ">u2"),
                ("Word7", ">u2"),
                ("Word8", ">u2"),
            ]
        )

    def _tm_832_(self: CCSDShdr) -> np.dtype:  # ApID = 0x340
        """Return data-type of MemDump packet."""
        if self.__hdr["length"] - 15 == 1:
            return np.dtype(
                [
                    ("hdr", self.__hdr.dtype),
                    ("Image_ID", "u1"),
                    ("_FillerByte", "u1"),
                    ("Address32", ">u4"),
                    ("Length", ">u4"),
                    ("Data", "u1"),
                ]
            )

        return np.dtype(
            [
                ("hdr", self.__hdr.dtype),
                ("Image_ID", "u1"),
                ("_FillerByte", "u1"),
                ("Address32", ">u4"),
                ("Length", ">u4"),
                (
                    "Data",
                    "u1",
                    (self.__hdr["length"] - 15),
                ),
            ]
        )

    def _tm_833_(self: CCSDShdr) -> np.dtype:  # ApID = 0x341
        """Return data-type of MemCheckRp packet."""
        return np.dtype(
            [
                ("hdr", self.__hdr.dtype),
                ("Image_ID", "u1"),
                ("_FillerByte", "u1"),
                ("Address32", ">u4"),
                ("Length", ">u4"),
                ("CheckSum", ">u4"),
            ]
        )

    def _tm_826_(self: CCSDShdr) -> np.dtype:  # ApID = 0x33A
        """Return data-type of MonListRp packet."""
        mon_dtype = np.dtype(
            [
                ("Mon1_EnSts", ">u2"),
                ("Mon1_ParID", ">u2"),
                ("Mon1_Int", "u1"),
                ("Mon1_NofSampl", "u1"),
                ("Mon1_CheckType", ">u2"),
                ("Mon1_LowOrExpCheckVal", ">u4"),
                ("Mon1_LowOrExpCheckRpId", ">u4"),
                ("Mon1_UppOrExpCheckVal", ">u4"),
                ("Mon1_UppOrExpCheckRpId", ">u4"),
            ]
        )
        num = (self.__hdr["length"] - 5) // mon_dtype.itemsize
        return np.dtype([("hdr", self.__hdr.dtype), ("Report", mon_dtype, (num,))])

    def _tm_827_(self: CCSDShdr) -> np.dtype:  # ApID = 0x33B
        """Return data-type of EvRpListRp packet."""
        return np.dtype(
            [
                ("hdr", self.__hdr.dtype),
                (
                    "Data",
                    "u1",
                    (self.__hdr["length"] - 5),
                ),
            ]
        )

    def _tm_828_(self: CCSDShdr) -> np.dtype:  # ApID = 0x33C
        """Return data-type of MpsTableRp packet."""
        return np.dtype(
            [
                ("hdr", self.__hdr.dtype),
                ("MPS_ID", "u1"),
                ("MPS_VER", "u1"),
                ("FTO", ">u2"),
                ("FTI", ">u2"),
                ("FTC", ">u2"),
                ("IMRO", ">u2"),
                ("IMRSA_A", ">u4"),
                ("IMRSA_B", ">u4"),
                ("IMRLEN", ">u4"),
                ("PKTLEN", ">u2"),
                ("TMRO", ">u2"),
                ("TMRI", ">u2"),
                ("IMDMODE", "u1"),
                ("I_LVDS", "u1"),
                ("_Filler1", ">u2"),
                ("_Filler2", ">u2"),
                ("_Filler3", ">u2"),
                ("DEM_RST", "u1"),
                ("DEM_CMV_CTRL", "u1"),
                ("COADD", "u1"),
                ("DEM_IGEN", "u1"),
                ("FRAME_MODE", "u1"),
                ("OUTPMODE", "u1"),
                ("BIN_TBL", ">u4"),
                ("COADD_BUF", ">u4"),
                ("COADD_RESA", ">u4"),
                ("COADD_RESB", ">u4"),
                ("FRAME_BUFA", ">u4"),
                ("FRAME_BUFB", ">u4"),
                ("LINE_ENA", ">u4"),
                ("NUMLIN", ">u2"),
                ("STR1", ">u2"),
                ("STR2", ">u2"),
                ("STR3", ">u2"),
                ("STR4", ">u2"),
                ("STR5", ">u2"),
                ("STR6", ">u2"),
                ("STR7", ">u2"),
                ("STR8", ">u2"),
                ("NUMLIN1", ">u2"),
                ("NUMLIN2", ">u2"),
                ("NUMLIN3", ">u2"),
                ("NUMLIN4", ">u2"),
                ("NUMLIN5", ">u2"),
                ("NUMLIN6", ">u2"),
                ("NUMLIN7", ">u2"),
                ("NUMLIN8", ">u2"),
                ("SUBS", ">u2"),
                ("SUBA", ">u2"),
                ("MONO", "u1"),
                ("IMFLP", "u1"),
                ("EXP_CTRL", "u1"),
                ("_FillerByte4", "u1"),
                ("EXP_TIME", ">u4"),
                ("EXP_STEP", ">u4"),
                ("EXP_KP1", ">u4"),
                ("EXP_KP2", ">u4"),
                ("NRSLOPE", "u1"),
                ("EXP_SEQ", "u1"),
                ("EXP_TIME2", ">u4"),
                ("EXP_STEP2", ">u4"),
                ("NUMFR", ">u2"),
                ("FOTLEN", "u1"),
                ("_FillerByte5", "u1"),
                ("ILVDSRCVR", "u1"),
                ("CALIB", "u1"),
                ("TRAINPTRN", ">u2"),
                ("CHENA", ">u4"),
                ("ICOL", "u1"),
                ("ICOLPR", "u1"),
                ("IADC", "u1"),
                ("IAMP", "u1"),
                ("VTFL1", "u1"),
                ("VTFL2", "u1"),
                ("VTFL3", "u1"),
                ("VRSTL", "u1"),
                ("VPRECH", "u1"),
                ("VREF", "u1"),
                ("VRAMP1", "u1"),
                ("VRAMP2", "u1"),
                ("OFFSET", ">u2"),
                ("PGAGAIN", "u1"),
                ("ADCGAIN", "u1"),
                ("TDIG1", "u1"),
                ("TDIG2", "u1"),
                ("BITMODE", "u1"),
                ("ADCRES", "u1"),
                ("PLLENA", "u1"),
                ("PLLINFRE", "u1"),
                ("PLLBYP", "u1"),
                ("PLLRATE", "u1"),
                ("PLLLOAD", "u1"),
                ("DETDUM", "u1"),
                ("BLACKCOL", "u1"),
                ("VBLACKSUN", "u1"),
                ("_Filler6", ">u4"),
                ("_Filler7", ">u4"),
            ]
        )

    def _tm_829_(self: CCSDShdr) -> np.dtype:  # ApID = 0x33D
        """Return data-type of ThemTableRp packet."""
        return np.dtype(
            [
                ("hdr", self.__hdr.dtype),
                ("HTR_1_IsEna", "u1"),
                ("HTR_1_AtcCorMan", "u1"),
                ("HTR_1_THMCH", "u1"),
                ("_FillerByte1", "u1"),
                ("HTR_1_ManOutput", ">u2"),
                ("HTR_1_ATC_SP", ">u4"),
                ("HTR_1_ATC_P", ">u4"),
                ("HTR_1_ATC_I", ">u4"),
                ("HTR_1_ATC_I_INIT", ">u4"),
                ("HTR_2_IsEna", "u1"),
                ("HTR_2_AtcCorMan", "u1"),
                ("HTR_2_THMCH", "u1"),
                ("_FillerByte2", "u1"),
                ("HTR_2_ManOutput", ">u2"),
                ("HTR_2_ATC_SP", ">u4"),
                ("HTR_2_ATC_P", ">u4"),
                ("HTR_2_ATC_I", ">u4"),
                ("HTR_2_ATC_I_INIT", ">u4"),
                ("HTR_3_IsEna", "u1"),
                ("HTR_3_AtcCorMan", "u1"),
                ("HTR_3_THMCH", "u1"),
                ("_FillerByte3", "u1"),
                ("HTR_3_ManOutput", ">u2"),
                ("HTR_3_ATC_SP", ">u4"),
                ("HTR_3_ATC_P", ">u4"),
                ("HTR_3_ATC_I", ">u4"),
                ("HTR_3_ATC_I_INIT", ">u4"),
                ("HTR_4_IsEna", "u1"),
                ("HTR_4_AtcCorMan", "u1"),
                ("HTR_4_THMCH", "u1"),
                ("_FillerByte4", "u1"),
                ("HTR_4_ManOutput", ">u2"),
                ("HTR_4_ATC_SP", ">u4"),
                ("HTR_4_ATC_P", ">u4"),
                ("HTR_4_ATC_I", ">u4"),
                ("HTR_4_ATC_I_INIT", ">u4"),
            ]
        )

    @property
    def hdr(self: CCSDShdr) -> np.ndarray:
        """Structured array holding the CCSDS header."""
        return self.__hdr

    @property
    def dtype(self: CCSDShdr) -> np.dtype:
        """Return numpy date-type of CCSDS headers."""
        return self.__dtype

    @property
    def version(self: CCSDShdr) -> int:
        """Return zero to indicate that this is a version 1 packet."""
        return (self.__hdr["type"] >> 13) & 0x7

    @property
    def type(self: CCSDShdr) -> int:
        """Return zero to indicate that this is a telemetry packet."""
        return (self.__hdr["type"] >> 12) & 0x1

    @property
    def apid(self: CCSDShdr) -> int:
        """Return ApID: an identifier for this telemetry packet.

        Notes
        -----
        SPEXone uses the following APIDs:

        - 0x320:  NomHk
        - 0x322:  DemHk
        - 0x331:  TcAccept
        - 0x332:  TcReject
        - 0x333:  TcExecute
        - 0x334:  TcFail
        - 0x335:  EventRp
        - 0x33A:  MonListRp
        - 0x33B:  EvRpListRp
        - 0x33C:  MpsTableRp
        - 0x33D:  ThermTableRp
        - 0x340:  MemDump
        - 0x341:  MemCheckRp
        - 0x350:  Science

        """
        return self.__hdr["type"] & 0x7FF

    @property
    def grouping_flag(self: CCSDShdr) -> int:
        """Data packages can be segmented.

        Note:
        ----
        This flag is encoded as::

         00 : continuation segment
         01 : first segment
         10 : last segment
         11 : unsegmented

        """
        return (self.__hdr["sequence"] >> 14) & 0x3

    @property
    def sequence(self: CCSDShdr) -> int:
        """Return the sequence counter.

        This counter is incremented with each consecutive packet of a
        particular ApID. This value will roll over to 0 after 0x3FF is reached.
        """
        return self.__hdr["sequence"] & 0x3FFF

    @property
    def packet_size(self: CCSDShdr) -> int:
        """Returns the CCSDS packet-length.

        Which is equal to the number of bytes in the Secondary header plus
        User Data minus 1.
        """
        return self.__hdr["length"]

    @property
    def tai_sec(self: CCSDShdr) -> int:
        """Seconds since 1958 (TAI)."""
        return self.__hdr["tai_sec"]

    @property
    def sub_sec(self: CCSDShdr) -> int:
        """Sub-seconds (1 / 2**16)."""
        return self.__hdr["sub_sec"]

    @property
    def data_dtype(self: CCSDShdr) -> np.dtype:
        """Return numpy data-type of CCSDS User Data."""
        method = getattr(self, f"_tm_{self.apid:d}_", None)
        return self._tm_raw_() if method is None else method()

    def tstamp(self: CCSDShdr, epoch: dt.datetime) -> dt.datetime:
        """Return time of the telemetry packet.

        Parameters
        ----------
        epoch :  datetime
           Provide the UTC epoch of the time (thus corrected for leap seconds)

        """
        return epoch + dt.timedelta(
            seconds=int(self.tai_sec),
            microseconds=100 * int(self.sub_sec / 65536 * 10000),
        )

    def read(self: CCSDShdr, file_format: str, buffer: bytes, offs: int = 0) -> None:
        """Read CCSDS primary and secondary headers from data.

        Parameters
        ----------
        file_format :  {'raw', 'dsb' or 'st3'}
           File format of this level 0 product
        buffer :  buffer_like
           Array of type (unsigned) byte.
        offs :  int, default=0
           Start reading the buffer from this offset (in bytes)

        Notes
        -----
        SPEXone level-0 file-formats:
        'raw'
           data has no file header and standard CCSDS packet headers
        'st3'
           data has no file header and ITOS + spacewire + CCSDS packet headers
        'dsb'
           data has a cFE file-header and spacewire + CCSDS packet headers

        """
        if file_format == "dsb":
            hdr_dtype = np.dtype(
                [
                    ("spacewire", "u1", (2,)),
                    ("type", ">u2"),
                    ("sequence", ">u2"),
                    ("length", ">u2"),
                    ("tai_sec", ">u4"),
                    ("sub_sec", ">u2"),
                ]
            )
        elif file_format == "raw":
            hdr_dtype = np.dtype(
                [
                    ("type", ">u2"),
                    ("sequence", ">u2"),
                    ("length", ">u2"),
                    ("tai_sec", ">u4"),
                    ("sub_sec", ">u2"),
                ]
            )
        elif file_format == "st3":
            hdr_dtype = np.dtype(
                [
                    ("itos_hdr", ">u2", (8,)),
                    ("spacewire", "u1", (2,)),
                    ("type", ">u2"),
                    ("sequence", ">u2"),
                    ("length", ">u2"),
                    ("tai_sec", ">u4"),
                    ("sub_sec", ">u2"),
                ]
            )
        else:
            raise ValueError("Unknown file_format, should be dsb, raw or st3")

        self.__hdr = np.frombuffer(buffer, count=1, offset=offs, dtype=hdr_dtype)[0]
        self.__dtype = hdr_dtype



#!/usr/bin/env python3
#
# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2019-2024 SRON - Netherlands Institute for Space Research
#    All Rights Reserved
#
# License:  BSD-3-Clause
"""Add OGSE information of a OCAL measurement to a SPEXone L1A product.

Possible OGSE information:

 * Reference diode (Ambient/GSFC polarized, radiometric): the reference diode
   is located in the ... or Grande (GSFC) integrated sphere.
 * Wavelength monitor (Ambient/GSFC polarized, radiometric, wavelength).
 * Wavelength of OPO laser (GSFC spectral-radiometry).
"""

from __future__ import annotations

import argparse
from pathlib import Path

from pyspex.ogse_db import (
    add_ogse_ref_diode,
    add_ogse_wav_mon,
    read_ref_diode,
    read_wav_mon,
)
from pyspex.ogse_dolp import gsfc_polarizer
from pyspex.ogse_grande import grande_spectrum
from pyspex.ogse_helios import helios_spectrum
from pyspex.ogse_laser import read_gse_excel

# - global parameters ------------------------------
DB_REF_DIODE = "ogse_db_ref_diode.nc"
DB_WAV_MON = "ogse_db_wave_mon.nc"


# - local functions --------------------------------
def create_ogse_db(args: argparse.Namespace) -> None:
    """Create databases for reference diode and/or Avantes fibre spectrometer."""
    if args.ref_diode:
        # read reference-diode data
        xds = read_ref_diode(args.ogse_dir, args.ref_diode, args.verbose)

        # create new database for reference-diode data
        xds.to_netcdf(
            args.ogse_dir / DB_REF_DIODE,
            mode="w",
            format="NETCDF4",
            group="/gse_data/ReferenceDiode",
        )

    if args.wav_mon:
        # read reference-diode data
        xds = read_wav_mon(args.ogse_dir, args.wav_mon, args.verbose)
        # create new database for reference-diode data
        xds.to_netcdf(
            args.ogse_dir / DB_WAV_MON,
            mode="w",
            format="NETCDF4",
            group="/gse_data/WaveMonitor",
        )


def write_ogse(args: argparse.Namespace) -> None:
    """Add OGSE data to a SPEXone level-1A product."""
    if args.ref_diode:
        add_ogse_ref_diode(args.ogse_dir / DB_REF_DIODE, args.l1a_file)

    if args.avantes:
        add_ogse_wav_mon(args.ogse_dir / DB_WAV_MON, args.l1a_file)

    if args.helios:
        xds = helios_spectrum()
        xds.to_netcdf(args.l1a_file, mode="a", group="/gse_data/ReferenceSpectrum")

    if args.grande:
        xds = gsfc_polarizer()
        xds.to_netcdf(args.l1a_file, mode="a", group="/gse_data/SpectralDolP")
        for n_lamps in (1, 2, 3, 5, 9):
            if args.l1a_file.name.find(f"-L{n_lamps:1d}_") > 0:
                xds = grande_spectrum(n_lamps)
                xds.to_netcdf(
                    args.l1a_file, mode="a", group="/gse_data/ReferenceSpectrum"
                )
                break

    if args.opo_laser:
        target_cwl = args.l1a_file.stem.split("_")[2].split("-")[-1]
        xds = read_gse_excel(args.ogse_dir, target_cwl)
        if xds is not None:
            xds.to_netcdf(args.l1a_file, mode="a", group="/gse_data/OPO_laser")


# - main function ----------------------------------
def main() -> None:
    """Add OGSE data to an existing SPEXone L1A product."""
    # parse command-line parameters
    parser = argparse.ArgumentParser()
    parser.add_argument("--verbose", action="store_true", help="be verbose")
    parser.add_argument(
        "--ogse_dir", default="Logs", type=Path, help="directory with OGSE data"
    )
    subparsers = parser.add_subparsers(help="sub-command help")
    parser_db = subparsers.add_parser("create_db", help="create new OGSE database")
    parser_db.add_argument(
        "--ref_diode", nargs="*", default=[], help="names of reference-diode files"
    )
    parser_db.add_argument(
        "--wav_mon",
        nargs="*",
        default=[],
        help="names of Avantes wavelength-monitor files",
    )
    parser_db.set_defaults(func=create_ogse_db)

    parser_wr = subparsers.add_parser(
        "add", help=("add OGSE information to a" " SPEXone Level-1A product")
    )
    parser_wr.add_argument(
        "--ref_diode",
        action="store_true",
        help="add reference-diode data from OGSE database",
    )
    parser_wr.add_argument(
        "--avantes",
        action="store_true",
        help=("add Avantes wavelength monitoring" "  from OGSE database"),
    )
    group_wr = parser_wr.add_mutually_exclusive_group()
    group_wr.add_argument(
        "--helios", action="store_true", help="add Helios reference spectrum"
    )
    group_wr.add_argument(
        "--grande", action="store_true", help="add Grande reference spectrum"
    )
    parser_wr.add_argument(
        "--opo_laser", action="store_true", help="add wavelength of OPO laser"
    )
    parser_wr.add_argument(
        "l1a_file", default=None, type=Path, help="SPEXone L1A product"
    )
    parser_wr.set_defaults(func=write_ogse)
    args = parser.parse_args()
    if args.verbose:
        print(args)

    # call whatever function was selected
    args.func(args)


# --------------------------------------------------
if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2019-2024 SRON - Netherlands Institute for Space Research
#    All Rights Reserved
#
# License:  BSD-3-Clause
"""Convert SPEXone OCAL measurement data to L1A format.

This script is obsoleted by spx1_level01a. Except for pdump files, which are
generated by the SPEXone EFM.
"""

from __future__ import annotations

import argparse
from datetime import UTC, datetime
from pathlib import Path

import numpy as np
from pyspex.ccsds_io import CCSDSio, hk_sec_of_day, img_sec_of_day
from pyspex.l1a_io import L1Aio
from pyspex.lib import pyspex_version
from pyspex.lib.tmtc_def import tmtc_dtype

# - global parameters ------------------------------
EPOCH = datetime(1970, 1, 1, tzinfo=UTC)


# - local functions --------------------------------
def get_l1a_name(msm_id: str) -> str:
    """Return name of SPEXone product for OCAL measurements (@SRON).

    Parameters
    ----------
    msm_id : string, optional
       Provide identifier for measurement, OCAL only

    Notes
    -----
    L1A file name format:
       SPX1_OCAL_<msm_id>[_YYYYMMDDTHHMMSS]_L1A_vvvvvvv.nc
    where
       msm_id is the measurement identifier
       YYYYMMDDTHHMMSS is time stamp of the first image in the file
       vvvvvvv is the git-hash string of the pyspex repository

    """
    # define string of sensing start as yyyymmddThhmmss
    return f"SPX1_OCAL_{msm_id}_L1A_{pyspex_version(githash=True)}.nc"


# - main function ----------------------------------
def main() -> None:
    """Create a SPEXone L1A product from data of as OCAL measurement.

    Examples
    --------
    Read CCSDS Level-0 files in current directory, write Level-1A in different
    directory::

    > spx1_ccsds2l1a.py --datapath L1A ./NomSciCal1_20220123T121801.676167

    Note that science data is read from ./NomSciCal1_20220123T121801.676167.?
    and ./NomSciCal1_20220123T121801.676167.??, and telemetry data is read
    from ./NomSciCal1_20220123T121801.676167_hk.?

    Read CCSDS Level-0 pdump-file generated by the EFM, write Level-1A in
    different directory::

    > spx1_ccsds2l1a.py --datapath L1A ./MPS_time_msg3_2020-11-02_20-20-04.200841.pdump

    """
    # parse command-line parameters
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description="Create SPEXone Level-1A product from CCSDS packages (L0)",
    )
    parser.add_argument("--debug", action="store_true")
    parser.add_argument("--verbose", action="store_true")
    parser.add_argument(
        "--select", default="all", choices=["all", "binned", "fullFrame"]
    )
    parser.add_argument(
        "--datapath",
        type=Path,
        default=Path("."),
        help="directory to store the Level-1A products",
    )
    parser.add_argument(
        "--pdump",
        action="store_true",
        help=("input files are CCSDS files" " generated by the EFM (with suffix)"),
    )
    parser.add_argument(
        "msmt_id",
        type=Path,
        help=(
            "name of the measurement without suffix"
            " (except for pdump-files). The names of the"
            " telemetry data are expected to be:"
            ' msmt_id+"_hk" (also without extension).'
        ),
    )
    args = parser.parse_args()
    if args.verbose:
        print(args)

    # Read Science packages
    data_dir = args.msmt_id.parent
    if not args.pdump:
        sci_files = sorted(data_dir.glob(args.msmt_id.name + ".[0-9]")) + sorted(
            data_dir.glob(args.msmt_id.name + ".?[0-9]")
        )
    else:
        sci_files = [args.msmt_id]

    packets = ()
    with CCSDSio(sci_files) as ccsds:
        while True:
            packet = ccsds.read_packet()
            # print(ccsds.fp.tell(), packet is None, ccsds.packet_length)
            if packet is None or ccsds.packet_length == 0:
                break

            if args.debug:
                print("[DEBUG]: ", ccsds)
            packets += (packet[0],)
    if args.verbose:
        print("[INFO]: number of CCSDS packets ", len(packets))

    # combine segmented packages
    science_tm = ccsds.science_tm(packets)
    del packets
    if args.verbose:
        print("[INFO]: number of Science images ", len(science_tm))

    num_packets = 0
    mps_length = 0
    mps_list = []
    for packet in science_tm:
        mps_id = packet["science_hk"]["MPS_ID"]
        mps_sz = packet["science_hk"]["IMRLEN"] // 2
        if mps_id == 0:
            continue

        if args.select == "fullFrame" and packet["science_hk"]["IMRLEN"] != 8388608:
            continue

        if args.select == "binned" and packet["science_hk"]["IMRLEN"] == 8388608:
            continue

        mps_length = max(mps_sz, mps_length)
        if mps_id not in mps_list:
            mps_list.append(mps_id)
        num_packets += 1
    if args.verbose:
        print(mps_length, mps_list)
        print("[INFO]: number of Science images ", num_packets)

    # read NomHK packages
    packets = ()
    if not args.pdump:
        hk_files = sorted(data_dir.glob(args.msmt_id.name + "_hk.[0-9]"))
        print(hk_files)

        with CCSDSio(hk_files) as ccsds:
            while True:
                packet = ccsds.read_packet()
                # print(ccsds.fp.tell(), packet is None, ccsds.packet_length)
                if packet is None or ccsds.packet_length == 0:
                    break

                if args.debug:
                    print("[DEBUG]: ", ccsds)
                packets += (packet[0],)
    if args.verbose:
        print("[INFO]: number of telemetry CCSDS packets ", len(packets))

    # select NomHK packages
    num_packets = 0
    nomhk_tm = ccsds.select_tm(packets, 0x320)
    for packet in nomhk_tm:
        if packet["nominal_hk"]["MPS_ID"] not in mps_list:
            continue
        num_packets += 1

    if args.debug or args.verbose:
        print("[INFO]: number of NomHK packages ", num_packets)

    # select DemHK packages
    if args.select == "all":
        demhk_tm = ccsds.select_tm(packets, 0x322)
        if args.debug or args.verbose:
            print("[INFO]: number of DemHK packages ", len(demhk_tm))

    if args.debug:
        return

    # Exit because we need Science data to create a valid L1A product
    if not science_tm:
        print("[WARNING]: no science data found, exit")
        return

    # extract timestaps, telemetry and image data from Science data
    img_sec = np.empty(len(science_tm), dtype="u4")
    img_subsec = np.empty(len(science_tm), dtype="u2")
    img_id = np.empty(len(science_tm), dtype="u4")
    img_hk = np.empty(len(science_tm), dtype=tmtc_dtype(0x350))
    img_data = []
    for ii, packet in enumerate(science_tm):
        if packet["science_hk"]["ICUSWVER"] > 0x123:
            img_sec[ii] = packet["icu_time"]["tai_sec"]
            img_subsec[ii] = packet["icu_time"]["sub_sec"]
        else:
            img_sec[ii] = packet["packet_header"]["tai_sec"]
            img_subsec[ii] = packet["packet_header"]["sub_sec"]
        img_id[ii] = packet["packet_header"]["sequence"] & 0x3FFF
        img_hk[ii] = packet["science_hk"]
        img_data.append(packet["image_data"])
    img_data = np.array(img_data)

    if np.all(img_hk["ICUSWVER"] == 0x123):
        # fix bug in sub-seconds
        us100 = np.round(10000 * img_subsec.astype(float) / 65536)
        buff = us100 + img_sec - 10000
        us100 = buff.astype("u8") % 10000
        img_subsec = ((us100 << 16) // 10000).astype("u2")

    ref_date, img_time = img_sec_of_day(img_sec, img_subsec, img_hk)

    # extract timestamps and telemetry of NomHK data
    if nomhk_tm:
        nomhk_sec = np.empty(len(nomhk_tm), dtype="u4")
        nomhk_subsec = np.empty(len(nomhk_tm), dtype="u2")
        nomhk_data = np.empty(len(nomhk_tm), dtype=tmtc_dtype(0x320))
        for ii, packet in enumerate(nomhk_tm):
            nomhk_sec[ii] = packet["packet_header"]["tai_sec"]
            nomhk_subsec[ii] = packet["packet_header"]["sub_sec"]
            nomhk_data[ii] = packet["nominal_hk"]

        if np.all(img_hk["ICUSWVER"] == 0x123):
            # fix bug in sub-seconds
            us100 = np.round(10000 * nomhk_subsec.astype(float) / 65536)
            buff = us100 + nomhk_sec - 10000
            us100 = buff.astype("u8") % 10000
            med = np.median(us100)
            indx = np.where(np.abs(us100 - med) > 1000)[0]
            us100[indx] = med
            nomhk_subsec = ((us100 << 16) // 10000).astype("u2")

    # if we have DemHK data then demhk_data should be equal in size with NomHK
    if demhk_tm:
        demhk_data = np.zeros(len(nomhk_tm), dtype=tmtc_dtype(0x322))
        for ii in range(min(len(nomhk_tm), len(demhk_tm))):
            demhk_data[ii] = demhk_tm[ii]["detector_hk"]

    # generate name of L1A product
    msm_id = args.msmt_id.name
    try:
        new_date = datetime.strptime(msm_id[-22:], "%y-%j-%H:%M:%S.%f").strftime(
            "%Y%m%dT%H%M%S.%f"
        )
    except ValueError:
        pass
    else:
        msm_id = msm_id[:-22] + new_date
    prod_name = get_l1a_name(msm_id)

    # pylint: disable=unsubscriptable-object
    n_frame = 1 if img_data.ndim == 1 else img_data.shape[0]
    n_sample = img_data.size if img_data.ndim == 1 else img_data.shape[1]
    if args.verbose:
        print(f"[INFO]: dimension of images [{n_frame},{n_sample}]")
    dims = {
        "number_of_images": n_frame,
        "samples_per_image": n_sample,
        "hk_packets": len(nomhk_tm),
        "SC_records": None,
    }

    # Generate L1A product
    with L1Aio(args.datapath / prod_name, dims=dims, ref_date=ref_date) as l1a:
        # write image data, detector telemetry and image attributes
        l1a.fill_science(img_data, img_hk, img_id)

        l1a.set_dset("/image_attributes/icu_time_sec", img_sec)
        l1a.set_attr(
            "valid_min", np.uint32(1577800000), ds_name="/image_attributes/icu_time_sec"
        )
        l1a.set_attr(
            "valid_max", np.uint32(1735700000), ds_name="/image_attributes/icu_time_sec"
        )
        l1a.set_attr(
            "units",
            "seconds since 1970-01-01 00:00:00",
            ds_name="/image_attributes/icu_time_sec",
        )
        l1a.set_dset("/image_attributes/icu_time_subsec", img_subsec)
        l1a.set_dset("/image_attributes/image_time", img_time)

        # write engineering data
        if nomhk_tm:
            l1a.fill_nomhk(nomhk_data)
            hk_time = hk_sec_of_day(nomhk_sec, nomhk_subsec, ref_date)
            l1a.set_dset("/engineering_data/HK_tlm_time", hk_time)

        if demhk_tm:
            l1a.set_dset("/engineering_data/DemHK_telemetry", demhk_data)

        # write global attributes
        l1a.fill_global_attrs(inflight=False)
        l1a.set_attr("input_files", [Path(x).name for x in sci_files])


# --------------------------------------------------
if __name__ == "__main__":
    main()



#
# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2023-2024 SRON - Netherlands Institute for Space Research
#    All Rights Reserved
#
# License:  BSD-3-Clause
#
"""`HKtlm`, class to read/access PACE/SPEXone telemetry data."""

from __future__ import annotations

__all__ = ["HKtlm"]

import datetime as dt
from copy import copy
from typing import TYPE_CHECKING

import numpy as np

from .ccsds_hdr import CCSDShdr
from .tlm_utils import convert_hk
from .tmtc_def import tmtc_dtype

if TYPE_CHECKING:
    import h5py
    import numpy.typing as npt


# - helper functions ------------------------
def subsec2musec(sub_sec: int) -> int:
    """Return subsec as microseconds."""
    return 100 * int(sub_sec / 65536 * 10000)


def mask2slice(mask: npt.NDArray[bool]) -> None | slice | tuple | npt.NDArray[bool]:
    """Try to slice (faster), instead of boolean indexing (slow)."""
    if np.all(~mask):
        return None
    if np.all(mask):
        return np.s_[:]  # read everything

    indx = np.nonzero(mask)[0]
    if np.all(np.diff(indx) == 1):
        # perform start-stop indexing
        return np.s_[indx[0] : indx[-1] + 1]

    # perform boolean indexing
    return mask


# - class HKtlm ----------------------------
class HKtlm:
    """Class to handle SPEXone housekeeping telemetry packets."""

    def __init__(self: HKtlm) -> None:
        """Initialize HKtlm object."""
        self.hdr: np.ndarray | None = None
        self.tlm: np.ndarray | None = None
        self.tstamp: list[dt.datetime, ...] | list = []
        self.events: list[np.ndarray, ...] | list = []

    def init_attrs(self: HKtlm) -> None:
        """Initialize class attributes."""
        self.hdr = None
        self.tlm = None
        self.tstamp = []
        self.events = []

    @property
    def size(self: HKtlm) -> int:
        """Return number of elements."""
        return 0 if self.tlm is None else len(self.tlm)

    def copy(self: HKtlm) -> HKtlm:
        """Return deep-copy of HKtlm object."""
        hkt = HKtlm()
        hkt.hdr = self.hdr.copy()
        hkt.tlm = self.tlm.copy()
        hkt.tstamp = copy(self.tstamp)
        hkt.events = copy(self.events)
        return hkt

    def sel(self: HKtlm, mask: np.NDArray[bool]) -> HKtlm:
        """Return subset of HKtlm object using a mask array."""
        hkt = HKtlm()
        hkt.hdr = self.hdr[mask]
        hkt.tlm = self.tlm[mask]
        hkt.tstamp = [x for x, y in zip(self.tstamp, mask, strict=True) if y]
        hkt.events = self.events.copy()
        return hkt

    def extract_l0_hk(self: HKtlm, ccsds_hk: tuple, epoch: dt.datetime) -> None:
        """Extract data from SPEXone level-0 housekeeping telemetry packets.

        Parameters
        ----------
        ccsds_hk :  tuple[np.ndarray, ...]
           SPEXone level-0 housekeeping telemetry packets
        epoch :  dt.datetime
           Epoch of the telemetry packets (1958 or 1970)

        """
        self.init_attrs()
        if not ccsds_hk:
            return

        self.hdr = np.empty(len(ccsds_hk), dtype=ccsds_hk[0]["hdr"].dtype)
        self.tlm = np.empty(len(ccsds_hk), dtype=tmtc_dtype(0x320))
        ii = 0
        for buf in ccsds_hk:
            ccsds_hdr = CCSDShdr(buf["hdr"][0])

            # Catch TcAccept, TcReject, TcExecute, TcFail and EventRp as events
            if ccsds_hdr.apid != 0x320 or buf["hdr"]["tai_sec"] < len(ccsds_hk):
                if 0x331 <= ccsds_hdr.apid <= 0x335:
                    self.events.append(buf)
                continue

            self.hdr[ii] = buf["hdr"]
            self.tlm[ii] = buf["hk"]
            self.tstamp.append(
                epoch
                + dt.timedelta(
                    seconds=int(buf["hdr"]["tai_sec"][0]),
                    microseconds=subsec2musec(buf["hdr"]["sub_sec"][0]),
                )
            )
            ii += 1

        self.hdr = self.hdr[:ii]
        self.tlm = self.tlm[:ii]

        # These values are originally stored in little-endian, but
        # Numpy does not accept a mix of little & big-endian values
        # in a structured array.
        self.tlm["HTR1_CALCPVAL"][:] = self.tlm["HTR1_CALCPVAL"].byteswap()
        self.tlm["HTR2_CALCPVAL"][:] = self.tlm["HTR2_CALCPVAL"].byteswap()
        self.tlm["HTR3_CALCPVAL"][:] = self.tlm["HTR3_CALCPVAL"].byteswap()
        self.tlm["HTR4_CALCPVAL"][:] = self.tlm["HTR4_CALCPVAL"].byteswap()
        self.tlm["HTR1_CALCIVAL"][:] = self.tlm["HTR1_CALCIVAL"].byteswap()
        self.tlm["HTR2_CALCIVAL"][:] = self.tlm["HTR2_CALCIVAL"].byteswap()
        self.tlm["HTR3_CALCIVAL"][:] = self.tlm["HTR3_CALCIVAL"].byteswap()
        self.tlm["HTR4_CALCIVAL"][:] = self.tlm["HTR4_CALCIVAL"].byteswap()

    def extract_l1a_hk(self: HKtlm, fid: h5py.File, mps_id: int | None) -> None:
        """Extract data from SPEXone level-1a housekeeping telemetry packets.

        Parameters
        ----------
        fid :  h5py.File
           File pointer to a SPEXone level-1a product
        mps_id : int, optional
           Select data performed with MPS equals 'mps_id'

        """
        self.init_attrs()

        # pylint: disable=no-member
        dset = fid["/engineering_data/NomHK_telemetry"]
        if mps_id is None:
            data_sel = np.s_[:]
        else:
            data_sel = mask2slice(dset.fields("MPS_ID")[:] == mps_id)
            if data_sel is None:
                return
        self.tlm = dset[data_sel]

        dset = fid["/engineering_data/HK_tlm_time"]
        ref_date = dset.attrs["units"].decode()[14:] + "+00:00"
        epoch = dt.datetime.fromisoformat(ref_date)
        for sec in dset[data_sel]:
            self.tstamp.append(epoch + dt.timedelta(seconds=sec))

    def convert(self: HKtlm, key: str) -> np.ndarray:
        """Convert telemetry parameter to physical units.

        Parameters
        ----------
        key :  str
           Name of telemetry parameter

        Returns
        -------
        np.ndarray

        """
        parm = key.upper()
        if parm in ("HTR1_POWER", "HTR2_POWER", "HTR3_POWER", "HTR4_POWER"):
            parm = parm.replace("_POWER", "_I")
        else:
            if parm not in self.tlm.dtype.names:
                raise KeyError(
                    f"Parameter: {parm} not found" f" in {self.tlm.dtype.names}"
                )

        raw_data = np.array([x[parm] for x in self.tlm])
        return convert_hk(key.upper(), raw_data)



#
# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2019-2024 SRON - Netherlands Institute for Space Research
#    All Rights Reserved
#
# License:  BSD-3-Clause
"""Read settings of the OPO laser used at NASA GSFC.

The OPO laser is used for ISRF and stray-light calibration measurements.
"""

from __future__ import annotations

__all__ = ["read_gse_excel"]

from pathlib import Path

import numpy as np
import xarray as xr
from openpyxl import load_workbook


# - main functions ----------------------------
def read_gse_excel(gse_dir: Path, target_cwl: str) -> xr.Dataset | None:
    """Return GSE info on central wavelength, line-width and laser radiance.

    Parameters
    ----------
    target_cwl: str
       Target laser wavelength as provided in the filename
    gse_dir: Path
       Name of folder with the Excel file

    """
    wbook = load_workbook(gse_dir / "SPEXOne_ALL_360-840nm.xlsx")
    wsheet = wbook.active
    actual_cwl = np.array([cel.value for cel in wsheet["E"][1:-1]])
    # print(actual_cwl)
    indx = np.argmin(np.abs(actual_cwl - float(target_cwl[:-2])))
    print(f"{target_cwl:>8s}: {actual_cwl[indx]:.3f} at index={indx:d}")
    if abs(float(target_cwl[:-2]) - actual_cwl[indx]) > 2:
        print("[WARNING]: no GSE information found")
        wbook.close()
        return None

    indx += 1
    wavelength = np.array([wsheet["E"][indx].value])
    xar_wv = xr.DataArray(
        wavelength,
        coords={"wavelength": wavelength},
        attrs={"long_name": "central wavelength", "units": "nm"},
    )
    xar_std = xr.DataArray(
        [wsheet["F"][indx].value],
        coords={"wavelength": wavelength},
        attrs={"long_name": "central wavelength [std]", "units": "nm"},
    )
    xar_lw = xr.DataArray(
        [wsheet["G"][indx].value],
        coords={"wavelength": wavelength},
        attrs={"long_name": "line width", "units": "nm"},
    )
    xar_sign = xr.DataArray(
        [wsheet["I"][indx].value],
        coords={"wavelength": wavelength},
        attrs={"long_name": "radiance", "units": "W/(m^2.sr)"},
    )
    wbook.close()
    return xr.Dataset(
        {
            "wavelength": xar_wv,
            "wv_std": xar_std,
            "linewidth": xar_lw,
            "signal": xar_sign,
        }
    )


def __test(l1a_file: str) -> None:
    """Small function to test this module."""
    # Create a netCDF4 file with the ISRF data in it
    gse_dir = Path("/data/richardh/SPEXone/GSFC")
    # xds = read_gse_excel(gse_dir, '465.4nm')
    # xds = read_gse_excel(gse_dir, '360nm', )
    xds = read_gse_excel(gse_dir, "840nm")
    xds.to_netcdf(l1a_file, mode="w", group="/gse_data/OPO_laser")


# --------------------------------------------------
if __name__ == "__main__":
    __test("test_netcdf.nc")



#
# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2019-2024 SRON - Netherlands Institute for Space Research
#    All Rights Reserved
#
# License:  BSD-3-Clause
#
"""Contains a collection of routines to access SPEXone Level-0 data.

Routines to read and write SPEXone CCSDS data:

   `dtype_tmtc`, `read_lv0_data`, `dump_hkt`, `dump_science`

Handy routines to convert CCSDS parameters:

   `hk_sec_of_day`, `img_sec_of_day`
"""

from __future__ import annotations

__all__ = ["CorruptPacketWarning", "dump_hkt", "dump_science", "read_lv0_data"]

import logging
import warnings
from pathlib import Path

import numpy as np

from .lib.ccsds_hdr import CCSDShdr
from .lib.tmtc_def import tmtc_dtype

# - local functions --------------------------------
module_logger = logging.getLogger("pyspex.lv0_lib")


def _cfe_header_(flname: Path) -> np.ndarray:
    """Read cFE file header (only for file_format='dsb')."""
    # define numpy data-type to read the cFE file-header
    dtype_cfe = np.dtype(
        [
            ("ContentType", "S4"),
            ("SubType", "S4"),
            ("FileHeaderLength", ">u4"),
            ("SpacecraftID", "S4"),
            ("ProcessorID", ">u4"),
            ("InstrumentID", "S4"),
            ("TimeSec", ">u4"),
            ("TimeSubSec", ">u4"),
            ("Filename", "S32"),
        ]
    )

    cfe_hdr = np.fromfile(flname, count=1, dtype=dtype_cfe)[0]
    module_logger.debug('content of cFE header "%s"', cfe_hdr)
    return cfe_hdr


def _fix_hk24_(sci_hk: np.ndarray) -> np.ndarray:
    """Correct 32-bit values in the Science HK.

    Which originate from 24-bit values in the detector register parameters.

    In addition::

    - copy the first 4 bytes of 'DET_CHENA' to 'DET_ILVDS'
    - parameter 'REG_BINNING_TABLE_START' was writen in little-endian

    """
    res = sci_hk.copy()
    if sci_hk["ICUSWVER"] < 0x129:
        res["REG_BINNING_TABLE_START"] = sci_hk["REG_BINNING_TABLE_START"].byteswap()

    res["DET_ILVDS"] = sci_hk["DET_CHENA"] & 0xF
    for key in [
        "TS1_DEM_N_T",
        "TS2_HOUSING_N_T",
        "TS3_RADIATOR_N_T",
        "TS4_DEM_R_T",
        "TS5_HOUSING_R_T",
        "TS6_RADIATOR_R_T",
        "LED1_ANODE_V",
        "LED1_CATH_V",
        "LED1_I",
        "LED2_ANODE_V",
        "LED2_CATH_V",
        "LED2_I",
        "ADC1_VCC",
        "ADC1_REF",
        "ADC1_T",
        "ADC2_VCC",
        "ADC2_REF",
        "ADC2_T",
        "DET_EXPTIME",
        "DET_EXPSTEP",
        "DET_KP1",
        "DET_KP2",
        "DET_EXPTIME2",
        "DET_EXPSTEP2",
        "DET_CHENA",
    ]:
        res[key] = sci_hk[key] >> 8

    return res


# - main function ----------------------------------
class CorruptPacketWarning(UserWarning):
    """Creating a custom warning."""


def read_lv0_data(
    file_list: list[Path], file_format: str, *, debug: bool = False
) -> tuple[tuple, tuple]:
    """Read level 0 data and return Science and telemetry data.

    Parameters
    ----------
    file_list : list of Path
       list of CCSDS files
    file_format : {'raw', 'st3', 'dsb'}
       type of CCSDS data
    debug : bool, default=False
       run in debug mode

    Returns
    -------
    tuple
         Contains all Science and TmTc CCSDS packages as numpy arrays,
         or None if called with debug is True

    """
    scihk_dtype = tmtc_dtype(0x350)
    icutm_dtype = np.dtype([("tai_sec", ">u4"), ("sub_sec", ">u2")])

    # read level 0 headers and CCSDS data of Science and TmTc data
    ccsds_sci = ()
    ccsds_hk = ()
    for flname in file_list:
        offs = 0
        if file_format == "dsb":
            cfe_hdr = _cfe_header_(flname)
            offs += cfe_hdr["FileHeaderLength"]

        buff_sci = ()  # Use chunking to speed-up memory allocation
        buff_hk = ()
        with open(flname, "rb") as fp:
            module_logger.debug('processing file "%s"', flname)

            # read CCSDS header and user data
            ccsds_data = fp.read()
            while offs < len(ccsds_data):
                try:
                    ccsds_hdr = CCSDShdr()
                    ccsds_hdr.read(file_format, ccsds_data, offs)
                    hdr_dtype = ccsds_hdr.dtype
                except ValueError as exc:
                    module_logger.warning('header read error with "%s".', exc)
                    break

                # check for data corruption (length > 0 and odd)
                if ccsds_hdr.apid != 0x340 and ccsds_hdr.packet_size % 2 == 0:
                    msg = (
                        "corrupted CCSDS packet detected:"
                        f" APID: {ccsds_hdr.apid}"
                        f", grouping_flag: {ccsds_hdr.grouping_flag}"
                        f", itemsize: {hdr_dtype.itemsize}"
                        f", packet_length: {ccsds_hdr.packet_size}"
                        f", file position: {offs}"
                    )
                    warnings.warn(msg, category=CorruptPacketWarning, stacklevel=1)
                    break

                if debug:
                    print(
                        ccsds_hdr.apid,
                        ccsds_hdr.grouping_flag,
                        hdr_dtype.itemsize,
                        ccsds_hdr.packet_size,
                        offs,
                    )
                    if not 0x320 <= ccsds_hdr.apid <= 0x350:
                        break

                    offs += hdr_dtype.itemsize + ccsds_hdr.packet_size - 5
                    continue

                # copy the full CCSDS package
                if ccsds_hdr.apid == 0x350:  # Science APID
                    nbytes = ccsds_hdr.packet_size - 5
                    if ccsds_hdr.grouping_flag == 1:
                        buff = np.empty(
                            1,
                            dtype=np.dtype(
                                [
                                    ("hdr", hdr_dtype),
                                    ("hk", scihk_dtype),
                                    ("icu_tm", icutm_dtype),
                                    ("frame", "O"),
                                ]
                            ),
                        )
                        buff["hdr"] = ccsds_hdr.hdr
                        offs += hdr_dtype.itemsize
                        buff["hk"] = _fix_hk24_(
                            np.frombuffer(
                                ccsds_data, count=1, offset=offs, dtype=scihk_dtype
                            )[0]
                        )
                        offs += scihk_dtype.itemsize
                        buff["icu_tm"] = np.frombuffer(
                            ccsds_data, count=1, offset=offs, dtype=icutm_dtype
                        )[0]
                        offs += icutm_dtype.itemsize
                        nbytes -= scihk_dtype.itemsize + icutm_dtype.itemsize
                    else:
                        buff = np.empty(
                            1, dtype=np.dtype([("hdr", hdr_dtype), ("frame", "O")])
                        )
                        buff["hdr"] = ccsds_hdr.hdr
                        offs += hdr_dtype.itemsize

                    buff["frame"][0] = np.frombuffer(
                        ccsds_data, count=nbytes // 2, offset=offs, dtype=">u2"
                    )
                    buff_sci += (buff.copy(),)
                    offs += nbytes
                elif 0x320 <= ccsds_hdr.apid < 0x350:  # other valid APIDs
                    dtype_tmtc = ccsds_hdr.data_dtype
                    buff = np.frombuffer(
                        ccsds_data, count=1, offset=offs, dtype=dtype_tmtc
                    )
                    buff_hk += (buff,)
                    offs += dtype_tmtc.itemsize
                else:
                    offs += hdr_dtype.itemsize + ccsds_hdr.packet_size - 5
            del ccsds_data

        ccsds_sci += buff_sci
        ccsds_hk += buff_hk

    module_logger.debug("number of Science packages %d", len(ccsds_sci))
    module_logger.debug("number of Engineering packages %d", len(ccsds_hk))

    return ccsds_sci, ccsds_hk


def dump_hkt(flname: str, ccsds_hk: tuple[np.ndarray, ...]) -> None:
    """Dump header info of the SPEXone housekeeping telemetry packets."""

    def msg_320(val: np.ndarray) -> str:
        return f" {val['ICUSWVER']:8x} {val['MPS_ID']:6d}"

    def msg_331(val: np.ndarray) -> str:
        return f" {-1:8x} {-1:6d} {val['TcSeqControl'][0]:12d}"

    def msg_332(val: np.ndarray) -> str:
        return (
            f" {-1:8x} {-1:6d} {val['TcSeqControl'][0]:12d}"
            f" {bin(val['TcRejectCode'][0])}"
            f" 0x{val['RejectParameter1'][0]:x}"
            f" 0x{val['RejectParameter2'][0]:x}"
        )

    def msg_333(val: np.ndarray) -> str:
        return f" {-1:8x} {-1:6d} {val['TcSeqControl'][0]:12d}"

    def msg_334(val: np.ndarray) -> str:
        return (
            f" {-1:8x} {-1:6d} {val['TcSeqControl'][0]:12d}"
            f" {bin(val['TcFailCode'][0])}"
            f" 0x{val['FailParameter1'][0]:x}"
            f" 0x{val['FailParameter2'][0]:x}"
        )

    def msg_335(val: np.ndarray) -> str:
        return (
            f" {-1:8x} {-1:6d} {bin(val['Event_ID'][0])}"
            f" {bin(val['Event_Sev'][0])}"
            f" 0x{val['Word1'][0]:x}"
            f" 0x{val['Word2'][0]:x}"
            f" 0x{val['Word3'][0]:x}"
            f" 0x{val['Word4'][0]:x}"
        )

    with Path(flname).open("w", encoding="ascii") as fp:
        fp.write(
            "APID Grouping Counter Length     TAI_SEC    SUB_SEC"
            " ICUSWVER MPS_ID TcSeqControl TcErrorCode\n"
        )
        for buf in ccsds_hk:
            ccsds_hdr = CCSDShdr(buf["hdr"][0])
            msg = (
                f"{ccsds_hdr.apid:4x} {ccsds_hdr.grouping_flag:8d}"
                f" {ccsds_hdr.sequence:7d} {ccsds_hdr.packet_size:6d}"
                f" {ccsds_hdr.tai_sec:11d} {ccsds_hdr.sub_sec:10d}"
            )

            if ccsds_hdr.apid == 0x320:
                msg_320(buf["hk"][0])
            else:
                method = {
                    0x331: msg_331,
                    0x332: msg_332,
                    0x333: msg_333,
                    0x334: msg_334,
                    0x335: msg_335,
                }.get(ccsds_hdr.apid, None)
                msg += "" if method is None else method(buf)
            fp.write(msg + "\n")


def dump_science(flname: str, ccsds_sci: tuple[np.ndarray, ...]) -> None:
    """Dump telemetry header info (Science)."""
    with Path(flname).open("w", encoding="ascii") as fp:
        fp.write(
            "APID Grouping Counter Length"
            " ICUSWVER MPS_ID  IMRLEN     ICU_SEC ICU_SUBSEC\n"
        )
        for segment in ccsds_sci:
            ccsds_hdr = CCSDShdr(segment["hdr"][0])
            if ccsds_hdr.grouping_flag == 1:
                nom_hk = segment["hk"]
                icu_tm = segment["icu_tm"]
                fp.write(
                    f"{ccsds_hdr.apid:4x}"
                    f" {ccsds_hdr.grouping_flag:8d}"
                    f" {ccsds_hdr.sequence:7d}"
                    f" {ccsds_hdr.packet_size:6d}"
                    f" {nom_hk['ICUSWVER'][0]:8x}"
                    f" {nom_hk['MPS_ID'][0]:6d}"
                    f" {nom_hk['IMRLEN'][0]:7d}"
                    f" {icu_tm['tai_sec'][0]:11d}"
                    f" {icu_tm['sub_sec'][0]:10d}\n"
                )
            else:
                fp.write(
                    f"{ccsds_hdr.apid:4x}"
                    f" {ccsds_hdr.grouping_flag:8d}"
                    f" {ccsds_hdr.sequence:7d}"
                    f" {ccsds_hdr.packet_size:6d}\n"
                )
